from __future__ import annotations
import abc
import logging
import typing
from typing import List
from PyQt5.QtWidgets import QWidget, QMessageBox

from pwspy import dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp import componentInterfaces
import pwspy_gui.PWSAnalysisApp.plugins



class CellSelectorPluginSupport:
    """A utility class to help manage CellSelectorPlugins"""
    def __init__(self, selector: componentInterfaces.CellSelector, pWidget: QWidget):
        pluginClasses = self._findPlugins()
        self._plugins: List[CellSelectorPlugin] = [clazz() for clazz in pluginClasses]
        for p in self._plugins:
            p.setContext(selector, pWidget)
        self._selector = selector
        self._parentWidget = pWidget

    def _findPlugins(self):
        """Scans the contents of pwspy_guiPWSAnalysisApp.plugins for any modules containing subclasses of CellSelectorPlugin.
        If someone wants to add a plugin without modifying this source code they can use namespace packages to make
        it seem as though their plugin module is in pwspy_guiPWSAnalysisApp.plugins"""
        import pkgutil, importlib, inspect
        iter_namespace = lambda pkg: pkgutil.iter_modules(pkg.__path__, pkg.__name__ + ".")  # Based on something I saw here https://packaging.python.org/guides/creating-and-discovering-plugins/#using-namespace-packages
        plugins = []

        for finder, name, ispkg in iter_namespace(pwspy_gui.PWSAnalysisApp.plugins):  # Find all submodules of the root module
            mod = importlib.import_module(name)
            clsmembers = inspect.getmembers(mod, lambda member: inspect.isclass(member) and not inspect.isabstract(member))  # Get all the classes that are defined in the module
            for name, cls in clsmembers:
                if issubclass(cls, CellSelectorPlugin):
                    plugins.append(cls)  # Add any class that implements the plugin base class
        return plugins

    # def registerPlugin(self, plugin: CellSelectorPlugin):  # this is from before we automatically scanned for plugins.
    #     self._plugins.append(plugin)
    #     plugin.setContext(self._selector)

    def getPlugins(self) -> typing.Sequence[CellSelectorPlugin]:
        return self._plugins

    def notifyCellSelectionChanged(self, cells: typing.Sequence[pwsdt.Acquisition]):
        for plugin in self._plugins:
            try:
                plugin.onCellsSelected(cells)
            except Exception as e:
                logging.getLogger(__name__).exception(e)
                QMessageBox.information(self._selector, "Plugin error", f"Error in `onCellsSelected` of plugin: {plugin.getName()}. See log.")

    def notifyReferenceSelectionChanged(self, cell: pwsdt.Acquisition):
        for plugin in self._plugins:
            try:
                plugin.onReferenceSelected(cell)
            except Exception as e:
                logging.getLogger(__name__).exception(e)
                QMessageBox.information(self._selector, "Plugin error", f"Error in `onReferenceSelected` of plugin: {plugin.getName()}. See log.")

    def notifyNewCellsLoaded(self, cells: List[pwsdt.Acquisition]):
        for plugin in self._plugins:
            try:
                plugin.onNewCellsLoaded(cells)
            except Exception as e:
                logging.getLogger(__name__).exception(e)
                QMessageBox.information(self._selector, "Plugin error", f"Error in `onNewCellsLoaded` of plugin: {plugin.getName()}. See log.")

    def notifyPluginSelected(self, plugin: CellSelectorPlugin):
        try:
            plugin.onPluginSelected()
        except Exception as e:
            logging.getLogger(__name__).exception(e)
            QMessageBox.information(self._selector, "Plugin error", f"Error in `onPluginSelected` of plugin: {plugin.getName()}. See log.")


class CellSelectorPlugin(metaclass=componentInterfaces.QABCMeta):
    """Base class for  a plugin that can extend the functionality of the `Cell Selector`.
    Implementions of this class should require no args for the constructor"""
    @abc.abstractmethod
    def setContext(self, selector: componentInterfaces.CellSelector, parentWidget: QWidget):
        """Set the CellSelector that this plugin is associated to. This should happen before anything else."""
        pass

    @abc.abstractmethod
    def onCellsSelected(self, cells: typing.Sequence[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that it has had new cells selected."""
        pass

    @abc.abstractmethod
    def onReferenceSelected(self, cell: pwsdt.Acquisition):
        """This method will be called when the CellSelector indicates that it has had a new reference selected."""
        pass

    @abc.abstractmethod
    def onNewCellsLoaded(self, cells: List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that new cells have been loaded to the selector."""
        pass

    @abc.abstractmethod
    def getName(self) -> str:
        """The name to refer to this plugin by."""
        pass

    @abc.abstractmethod
    def onPluginSelected(self):
        """This method will be called when the plugin is activated."""
        pass

    @abc.abstractmethod
    def additionalColumnNames(self) -> typing.Sequence[str]:
        """The header names for each column."""
        pass

    @abc.abstractmethod
    def getTableWidgets(self, acq: pwsdt.Acquisition) -> typing.Sequence[QWidget]:
        """provide a widget for each additional column to represent `acq`"""
        pass

if __name__ == '__main__':
    s = CellSelectorPluginSupport(None)
    d = s._findPlugins()
    a = 1



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union

from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget, QCheckBox, QFrame, QScrollArea, QGridLayout, QSizePolicy, QLayout


class CollapsibleSection(QWidget):
    stateChanged = QtCore.pyqtSignal(bool)

    def __init__(self, title, animationDuration, parent: QWidget):
        super().__init__(parent)
        self._animationDuration = animationDuration
        self._toggleButton = QCheckBox(title, self)
        headerLine = QFrame(self)
        self._toggleAnimation = QtCore.QParallelAnimationGroup(self)
        self._contentArea = QScrollArea(self)
        mainLayout = QGridLayout(self)

        self._toggleButton.setCheckable(True)
        self._toggleButton.setChecked(True)

        headerLine.setFrameShape(QFrame.HLine)
        headerLine.setFrameShadow(QFrame.Sunken)
        headerLine.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        self._contentArea.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        # start out collapsed
        self._contentArea.setMaximumHeight(0)
        self._contentArea.setMinimumHeight(0)

        # let the entire widget grow and shrink with its content
        self._toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self, b"minimumHeight"))
        self._toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self, b"maximumHeight"))
        self._toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self._contentArea, b"maximumHeight"))

        mainLayout.setVerticalSpacing(0)
        mainLayout.setContentsMargins(0, 0, 0, 0)

        mainLayout.addWidget(self._toggleButton, 0, 0, 1, 1, QtCore.Qt.AlignLeft)

        mainLayout.addWidget(headerLine, 1, 2, 1, 1)

        mainLayout.addWidget(self._contentArea, 1, 0, 1, 3)

        self.setLayout(mainLayout)
        self.setLayout = self._setLayout

        self._toggleButton.toggled.connect(
            lambda checked:
            [self._toggleAnimation.setDirection(
                    QtCore.QAbstractAnimation.Forward if not checked else QtCore.QAbstractAnimation.Backward),
            self._toggleAnimation.start()])
        self._toggleAnimation.finished.connect(lambda: self.stateChanged.emit(self._toggleButton.isChecked()))

    def checkState(self):
        return self._toggleButton.checkState()

    def setCheckState(self, state: Union[bool, int]):
        self._toggleButton.setCheckState(state)

    def _setLayout(self, contentLayout: QLayout):
        self._contentArea.setLayout(contentLayout)
        collapsedHeight = self.sizeHint().height() - self._contentArea.maximumHeight()
        contentHeight = contentLayout.sizeHint().height()
        for i in range(self._toggleAnimation.animationCount() - 1):
            SectionAnimation = self._toggleAnimation.animationAt(i)
            SectionAnimation.setDuration(self._animationDuration)
            SectionAnimation.setStartValue(collapsedHeight)
            SectionAnimation.setEndValue(collapsedHeight + contentHeight)

        contentAnimation = self._toggleAnimation.animationAt(self._toggleAnimation.animationCount() - 1)
        contentAnimation.setDuration(self._animationDuration)
        contentAnimation.setStartValue(0)
        contentAnimation.setEndValue(contentHeight)


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from PyQt5 import QtGui, QtCore
from PyQt5.QtWidgets import QWidget, QWIDGETSIZE_MAX


class AspectRatioWidget(QWidget):
    def __init__(self, aspect: float, parent: QWidget = None):
        super().__init__(parent)
        self._aspect = aspect

    def resizeEvent(self, event: QtGui.QResizeEvent):
        w, h = event.size().width(), event.size().height()
        self._resize(w, h)

    def _resize(self, width, height):
        newHeight = width / self._aspect #The ideal height based on the new commanded width
        newWidth = height * self._aspect #the ideal width based on the new commanded height
        #Now determine which of the new dimensions to use.
        if width > newWidth:
            self.setMaximumWidth(newWidth)
            self.setMaximumHeight(QWIDGETSIZE_MAX)
        else:
            self.setMaximumHeight(newHeight)
            self.setMaximumWidth(QWIDGETSIZE_MAX)

    def setAspect(self, aspect: float):
        self._aspect = aspect
        self._resize(self.width(), self.height())



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Wed Feb 13 18:04:57 2019

@author: Nick Anthony
"""
from __future__ import annotations
import os
from glob import glob
from pwspy_gui import resources
from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import (QGridLayout, QDialog,
                             QLineEdit, QPushButton, QFileDialog, QCheckBox,
                             QMessageBox, QWidget, QVBoxLayout, QTreeWidget, QTreeWidgetItem, QApplication)

import typing

from pwspy_gui.PWSAnalysisApp._dockWidgets.ResultsTableDock import ConglomerateCompilerResults

if typing.TYPE_CHECKING:
    from typing import Optional, List, Tuple
    from pwspy.dataTypes import ICMetaData
    from pwspy.analysis.compilation import PWSRoiCompilationResults
    from pwspy.analysis.pws import PWSAnalysisSettings
    from pwspy.analysis.warnings import AnalysisWarning


class WorkingDirDialog(QDialog):
    directoryChanged = QtCore.pyqtSignal(str, bool)

    def __init__(self, parent=None):
        super().__init__(parent,
                         QtCore.Qt.WindowTitleHint | QtCore.Qt.WindowSystemMenuHint | QtCore.Qt.WindowCloseButtonHint)  # Construct without a question mark button
        self.setWindowTitle("Working Directory")
        layout = QGridLayout()
        self.setModal(True)
        self.textLine = QLineEdit()
        self.browseButton = QPushButton(QtGui.QIcon(os.path.join(resources, 'folder.svg')), '')
        self.scanButton = QPushButton('Scan!')
        self.recursiveCheckbox = QCheckBox("recursively scan subfolders")
        self.recursiveCheckbox.setChecked(True)
        layout.addWidget(self.textLine, 0, 0, 1, 4)
        layout.addWidget(self.browseButton, 0, 5, 1, 1)
        layout.addWidget(self.recursiveCheckbox, 1, 0, 1, 2)
        layout.addWidget(self.scanButton, 1, 2, 1, 1)
        self.setLayout(layout)
        self.setFixedSize(400, 75)
        self.scanButton.released.connect(self._scanButtonPushed)
        self.browseButton.released.connect(self.browseFile)
        self.directory = os.path.expanduser('~')

    def _scanButtonPushed(self):
        self.directory = self.textLine.text()
        recursive = self.recursiveCheckbox.checkState() != 0
        self.directoryChanged.emit(self.directory, recursive)
        self.accept()

    def browseFile(self):
        _ = QFileDialog.getExistingDirectory(self, 'Working Directory', self.directory)
        if _ != '':
            self.directory = _
            self.textLine.setText(self.directory)

    def show(self):
        super().show()
        self.browseFile()


class AnalysisSummaryDisplay(QDialog):
    def __init__(self, parent: Optional[QWidget], warnings: List[Tuple[List[AnalysisWarning], ICMetaData]], analysisName: str = '', analysisSettings: PWSAnalysisSettings = None):
        super().__init__(parent=parent)
        self.analysisName = analysisName
        self.analysisSettings = analysisSettings
        layout = QVBoxLayout()
        self.settingsButton = QPushButton("Settings", self)
        self.settingsButton.released.connect(self._displaySettings)
        self.warnList = QTreeWidget(self)
        self.warnList.setHeaderHidden(True)
        layout.addWidget(self.settingsButton)
        layout.addWidget(self.warnList)
        self.setLayout(layout)
        self._addWarnings(warnings)
        self.setWindowTitle(f"Analysis Summary: {analysisName}")
        self.show()

    def _addWarnings(self, warnings: List[Tuple[List[AnalysisWarning],ICMetaData]]):
         for cellWarns, cell in warnings:
            item = QTreeWidgetItem(self.warnList)
            item.setText(0, cell.filePath)
            for warn in cellWarns:
                subItem = QTreeWidgetItem(item)
                subItem.setText(0, warn.shortMsg)
                subItem.setToolTip(0, warn.longMsg)

    def clearWarnings(self):
        self.warnList.clear()

    def _displaySettings(self):
        if self.analysisSettings is not None:
            msgBox = QMessageBox.information(self, self.analysisName, self.analysisSettings.toJsonString())

if __name__ == '__main__':
    _ = WorkingDirDialog()
    _.show()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import sys, os
import typing
from numbers import Number
from typing import Tuple
import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets

__all__ = ['QRangeSlider']

from PyQt5.QtGui import QColor

from PyQt5.QtWidgets import QHBoxLayout, QWidget, QGridLayout, QSplitter

DEFAULT_CSS= """
QRangeSlider * {
    border: 0px;
    padding: 0px;
}

QRangeSlider #Head {
    background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #222, stop:1 #333);
}
QRangeSlider #Span:active {
    background: #339099;
}
QRangeSlider #Tail {
    background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #222, stop:1 #333);
}
QRangeSlider > QSplitter::handle {
    background: #d42a04;
}
QRangeSlider > QSplitter::handle:vertical {
    height: 4px;
}
QRangeSlider > QSplitter::handle:pressed {
    background: #691401;
}
"""


def scale(val: Number, src: Tuple[Number, Number], dst: Tuple[Number, Number]):
    """src is a tuple containing the original minimum and maximum values.
    dst is a tuple containing the new min and max values.
    This function will return the value of val scaled from src to dst"""
    if src[1] == src[0]: #These cases will result in a nan value.
        return (dst[1] + dst[0])/2
    elif dst[0] == dst[1]:
        return dst[0]
    else:
        return float(((val - src[0]) / float(src[1]-src[0])) * (dst[1]-dst[0]) + dst[0])


class Element(QtWidgets.QWidget):
    def __init__(self, main):
        super().__init__()
        self.main = main
        self._textColor = QtGui.QColor(125, 125, 125)

    def textColor(self) -> QtGui.QColor:
        return self._textColor

    def setTextColor(self, color: typing.Union[int, typing.Tuple[int, int, int]]):
        if type(color) == tuple and len(color) == 3:
            self._textColor = QtGui.QColor(color[0], color[1], color[2])
        elif type(color) == int:
            self._textColor = QtGui.QColor(color, color, color)

    def paintEvent(self, event):
        qp = QtGui.QPainter(self)
        if self.main.drawValues():
            self.drawText(event, qp)

    def drawText(self, event, painter):
        pass #Override this to draw text.

    def setStyleSheet(self, styleSheet: str) -> None:
        super().setStyleSheet(styleSheet)
        self.setAutoFillBackground(True) # For some reason if this isn't done then setting the stylesheet turns everything gray.


def numFormat(num: Number) -> str:
    num = np.abs(num)
    if num < 1 or num>=1000:
        return f"{num:.2E}"
    else:
        return f"{num:.2f}"


class Head(Element):
    def __init__(self, main):
        super(Head, self).__init__(main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont('Arial', 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, numFormat(self.main.min()))


class Tail(Element):
    def __init__(self, main):
        super(Tail, self).__init__(main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont('Arial', 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignRight, numFormat(self.main.max()))


class Handle(Element):
    def __init__(self, main):
        super(Handle, self).__init__(main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont('Arial', 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, numFormat(self.main.start()))
        qp.drawText(event.rect(), QtCore.Qt.AlignRight, numFormat(self.main.end()))

    def mouseMoveEvent(self, event):
        event.accept()
        mx = event.globalX()
        _mx = getattr(self, '__mx', None)
        if not _mx:
            setattr(self, '__mx', mx)
            dx = 0
        else:
            dx = mx - _mx
        setattr(self, '__mx', mx)
        if dx == 0:
            event.ignore()
            return
        elif dx > 0:
            dx = 1
        elif dx < 0:
            dx = -1
        s = self.main.start() + dx
        e = self.main.end() + dx
        if s >= self.main.min() and e <= self.main.max():
            self.main.setRange(s, e)


class QRangeSlider(QtWidgets.QWidget):
    endValueChanged = QtCore.pyqtSignal(int)
    maxValueChanged = QtCore.pyqtSignal(int)
    minValueChanged = QtCore.pyqtSignal(int)
    startValueChanged = QtCore.pyqtSignal(int)

    _SPLIT_START = 1
    _SPLIT_END = 2

    def __init__(self, parent=None):
        super(QRangeSlider, self).__init__(parent)
        # setup Ui
        self.resize(300, 30)
        self.gridLayout = QtWidgets.QGridLayout(self)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self._splitter = QtWidgets.QSplitter(self)
        self._splitter.setMinimumSize(QtCore.QSize(0, 0))
        self._splitter.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self._splitter.setOrientation(QtCore.Qt.Horizontal)
        self.gridLayout.addWidget(self._splitter, 0, 0, 1, 1)
        self.head = Head(main=self)  #The order that these are added matters.
        self.head.setObjectName("Head") # These names are used by the stylesheet
        self.handle = Handle(main=self)
        self.handle.setObjectName("Span")
        self.tail = Tail(main=self)
        self.tail.setObjectName("Tail")
        self.setMinimumHeight(15)

        self._splitter.addWidget(self.head)
        self._splitter.addWidget(self.handle)
        self._splitter.addWidget(self.tail)

        self.setStyleSheet(DEFAULT_CSS)

        self.setMouseTracking(False)  # Don't fire mouse events unless a button is clicked.
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self.handle.setTextColor((150, 255, 150))
        self._setMin(0)
        self._setMax(99)
        self.setStart(0)
        self.setEnd(99)
        self.setDrawValues(True)

    def showEvent(self, evt):
        self.head.setAutoFillBackground(True) #IDK why but the colors from the stylesheet won't fill in if we don't do this.
        self.tail.setAutoFillBackground(True)
        self.handle.setAutoFillBackground(True)
        super().showEvent(evt)

    def min(self):
        return getattr(self, '__min', None)

    def max(self):
        return getattr(self, '__max', None)

    def _setMin(self, value):
        setattr(self, '__min', value)
        self.minValueChanged.emit(value)

    def setMin(self, value):
        self._setMin(value)
        if self.start() < value:
            self._setStart(value)

    def _setMax(self, value):
        setattr(self, '__max', value)
        self.maxValueChanged.emit(value)

    def setMax(self, value):
        self._setMax(value)
        if self.end() > value:
            self._setEnd(value)

    def start(self):
        return getattr(self, '__start', None)

    def end(self):
        return getattr(self, '__end', None)

    def _setStart(self, value):
        setattr(self, '__start', value)
        self.startValueChanged.emit(value)

    def setStart(self, value):
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_START)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setStart(value)

    def _setEnd(self, value):
        setattr(self, '__end', value)
        self.endValueChanged.emit(value)

    def setEnd(self, value):
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_END)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setEnd(value)

    def drawValues(self):
        return getattr(self, '__drawValues', None)

    def setDrawValues(self, draw):
        setattr(self, '__drawValues', draw)

    def getRange(self):
        return self.start(), self.end()

    def setRange(self, start, end):
        self.setStart(start)
        self.setEnd(end)

    def keyPressEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Left:
            s = self.start()-1
            e = self.end()-1
        elif key == QtCore.Qt.Key_Right:
            s = self.start()+1
            e = self.end()+1
        else:
            event.ignore()
            return
        event.accept()
        if s >= self.min() and e <= self.max():
            self.setRange(s, e)

    def setBackgroundStyle(self, style):
        self.tail.setStyleSheet(style)
        self.head.setStyleSheet(style)

    def _valueToPos(self, value):
        s = scale(value, (self.min(), self.max()), (0, self.width()))
        return int(s)

    def _posToValue(self, xpos):
        return scale(xpos, (0, self.width()), (self.min(), self.max()))

    def _handleMoveSplitter(self, xpos, index):
        hw = self._splitter.handleWidth()
        def _lockWidth(widget):
            width = widget.size().width()
            widget.setMinimumWidth(width)
            widget.setMaximumWidth(width)
        def _unlockWidth(widget):
            widget.setMinimumWidth(0)
            widget.setMaximumWidth(16777215)
        v = self._posToValue(xpos)
        if index == self._SPLIT_START:
            _lockWidth(self.tail)
            if v >= self.end():
                return
            offset = -20
            w = xpos + offset
            self._setStart(v)
        elif index == self._SPLIT_END:
            _lockWidth(self.head)
            if v <= self.start():
                return
            offset = -40
            w = self.width() - xpos + offset
            self._setEnd(v)
        _unlockWidth(self.tail)
        _unlockWidth(self.head)
        _unlockWidth(self.handle)

if __name__ == '__main__':

    app = QtWidgets.QApplication(sys.argv)
    rs = QRangeSlider()
    rs.show()

    rs.setMax(100)
    rs.setMin(0.015)
    rs.setRange(.017, .50)
    rs.setBackgroundStyle('background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #222, stop:1 #333);')
    rs.handle.setStyleSheet('background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #282, stop:1 #393);')
    app.exec_()


import abc
import os
from PyQt5.QtCore import QObject
from cachetools import cachedmethod, LRUCache
from pwspy import dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp.componentInterfaces import ROIManager


class _DefaultROIManager(ROIManager, QObject):
    def __init__(self, parent: QObject = None):
        super().__init__(parent=parent)
        self._cache = LRUCache(maxsize=2048)  # Store this many ROIs at once

    @staticmethod
    def _getCacheKey(roiFile: pwsdt.RoiFile):
        return os.path.split(roiFile.filePath)[0], roiFile.name, roiFile.number

    def removeRoi(self, roiFile: pwsdt.RoiFile):
        self._cache.pop(self._getCacheKey(roiFile))
        roiFile.delete()
        self.roiRemoved.emit(roiFile)

    def updateRoi(self, roiFile: pwsdt.RoiFile, roi: pwsdt.Roi):
        roiFile.update(roi)
        self._cache[self._getCacheKey(roiFile)] = roiFile
        self.roiUpdated.emit(roiFile)

    def createRoi(self, acq: pwsdt.Acquisition, roi: pwsdt.Roi, roiName: str, roiNumber: int, overwrite: bool = False) -> pwsdt.RoiFile:
        roiFile = acq.saveRoi(roiName, roiNumber, roi, overwrite=overwrite)
        self._cache[self._getCacheKey(roiFile)] = roiFile
        self.roiCreated.emit(roiFile, overwrite)
        return roiFile

    @cachedmethod(lambda self: self._cache, key=lambda acq, roiName, roiNum: (acq.filePath, roiName, roiNum))  # Cache results
    def getROI(self, acq: pwsdt.Acquisition, roiName: str, roiNum: int) -> pwsdt.RoiFile:
        return acq.loadRoi(roiName, roiNum)

    def close(self):
        self._cache.clear()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from typing import List, Tuple, Optional
from pwspy.analysis import warnings
from pwspy.analysis.dynamics import DynamicsAnalysisResults
from pwspy.analysis.pws import PWSAnalysisResults
from pwspy.dataTypes import Roi
from pwspy.analysis.compilation import (DynamicsRoiCompiler, DynamicsCompilerSettings, DynamicsRoiCompilationResults,
                                        PWSRoiCompiler, PWSCompilerSettings, PWSRoiCompilationResults,
                                        GenericRoiCompiler, GenericCompilerSettings, GenericRoiCompilationResults)
from typing import NamedTuple
"""These utility classes are used to conveniently treat analysis objects of different types that belong together as a single object."""


class ConglomerateCompilerSettings(NamedTuple):
    pws: PWSCompilerSettings
    dyn: DynamicsCompilerSettings
    generic: GenericCompilerSettings


class ConglomerateCompilerResults(NamedTuple):
    pws: PWSRoiCompilationResults
    dyn: DynamicsRoiCompilationResults
    generic: GenericRoiCompilationResults


class ConglomerateAnalysisResults(NamedTuple):
    pws: Optional[PWSAnalysisResults]
    dyn: Optional[DynamicsAnalysisResults]


class ConglomerateCompiler:
    """
    This convenience class combines the actions of the compilers defined in `pwspy.analysis.compilation` into a single object.
    """
    def __init__(self, settings: ConglomerateCompilerSettings):
        self.settings = settings
        self.pws = PWSRoiCompiler(self.settings.pws)
        self.dyn = DynamicsRoiCompiler(self.settings.dyn)
        self.generic = GenericRoiCompiler(self.settings.generic)

    def run(self, results: ConglomerateAnalysisResults, roiFile: RoiFile) -> Tuple[ConglomerateCompilerResults, List[warnings.AnalysisWarning]]:
        if results.pws is not None:
            pwsResults, pwsWarnings = self.pws.run(results.pws, roiFile)
        else:
            pwsResults, pwsWarnings = None, []
        if results.dyn is not None:
            dynResults, dynWarnings = self.dyn.run(results.dyn, roiFile)
        else:
            dynResults, dynWarnings = None, []
        genResults = self.generic.run(roiFile)
        return ConglomerateCompilerResults(pwsResults, dynResults, genResults), pwsWarnings + dynWarnings



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.
import typing as t_
import json
import os
from json import JSONDecodeError

import logging
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import QPushButton, QTableWidgetItem, QTableWidget, QAbstractItemView, QMenu, QWidget, QMessageBox, \
    QInputDialog, QHeaderView

from pwspy_gui.PWSAnalysisApp.sharedWidgets import ScrollableMessageBox
from pwspy.dataTypes import Acquisition, ICMetaData, DynMetaData

from pwspy_gui.PWSAnalysisApp.sharedWidgets.dictDisplayTree import DictDisplayTreeDialog
from pwspy_gui.PWSAnalysisApp.sharedWidgets.tables import NumberTableWidgetItem
if t_.TYPE_CHECKING:
    from pwspy.analysis.pws import PWSAnalysisResults
    from pwspy.analysis.dynamics import DynamicsAnalysisResults


def evalToolTip(cls: t_.Type[QWidget], method):
    """Given a QWidget and a function that returns a string, this decorator returns a modified class that will evaluate
    the function each time the tooltip is requested."""
    class newClass(cls):
        def event(self, e: QtCore.QEvent):
            if e.type() == QtCore.QEvent.ToolTip:
                self.setToolTip(method())
            return super().event(e)
    return newClass


class PreferencesMetadata:
    def __init__(self, filePath: str, invalid: bool = False, reference: bool = False):
        self.filePath = filePath
        self._invalid = invalid
        self._reference = reference
        self._stale: bool = True  # Keeps track of if this object is in sync with the file. False means no saving is needed.

    def close(self):
        if self._stale:
            self._toJson()  # Save if needed.

    def __del__(self):  # This doesn't always work but it doesn't hurt to try
        self.close()

    @property
    def reference(self): return self._reference

    @reference.setter
    def reference(self, ref: bool):
        if self._reference != ref:
            self._reference = ref
            self._stale = True

    @property
    def invalid(self): return self._invalid

    @invalid.setter
    def invalid(self, inv: bool):
        if self._invalid != inv:
            self._invalid = inv
            self._stale = True


    def _toJson(self):
        d = {'invalid': self._invalid, 'reference': self._reference}
        with open(self.filePath, 'w') as f:
            json.dump(d, f)
        self._stale = False

    @classmethod
    def fromJson(cls, filePath: str):
        with open(filePath, 'r') as f:
            d = json.load(f)
        md = PreferencesMetadata(filePath, invalid=d['invalid'], reference=d['reference'])
        md._stale = False
        return md


class CellTableWidgetItem:
    """Represents a single row of the CellTableWidget and corresponds to a single PWS acquisition."""
    def __init__(self, acq: Acquisition, label: str, num: int, additionalWidgets: t_.Sequence[QWidget] = None):
        self.Acquisition = acq
        self.num = num
        self.path = label
        self.pluginWidgets = [] if additionalWidgets is None else additionalWidgets
        self.notesButton = evalToolTip(QPushButton, acq.getNotes)("Open")
        self.notesButton.setFixedSize(40, 30)
        self.pathLabel = QTableWidgetItem(self.path)
        self.pathLabel.setToolTip(acq.idTag)
        self.numLabel = NumberTableWidgetItem(num)
        self.numLabel.setToolTip(acq.idTag)
        self.roiLabel = NumberTableWidgetItem(0)
        self.anLabel = NumberTableWidgetItem(0)
        self.notesButton.released.connect(self.Acquisition.editNotes)
        self.pLabel = QTableWidgetItem()
        self.pLabel.setToolTip("Indicates if PWS measurement is present")
        self.dLabel = QTableWidgetItem()
        self.dLabel.setToolTip("Indicates if Dynamics measurement is present")
        self.fLabel = QTableWidgetItem()
        self.fLabel.setToolTip("Indicates if Fluorescence measurement is present")
        for i in [self.pLabel, self.dLabel, self.fLabel]:
            i.setTextAlignment(QtCore.Qt.AlignCenter)
        for i in [self.pathLabel, self.pLabel, self.dLabel, self.fLabel]:  # Make uneditable
            i.setFlags(i.flags() ^ QtCore.Qt.ItemIsEditable)
        for metadata, label in [(self.Acquisition.pws, self.pLabel), (self.Acquisition.dynamics, self.dLabel)]:
            if metadata is not None:
                label.setText('Y')
                label.setBackground(QtCore.Qt.darkGreen)
                label.setToolTip(metadata.idTag)
            else:
                label.setText('N')
                label.setBackground(QtCore.Qt.white)
        if len(self.Acquisition.fluorescence) != 0: self.fLabel.setText('Y'); self.fLabel.setBackground(QtCore.Qt.darkGreen)
        else: self.fLabel.setText('N'); self.fLabel.setBackground(QtCore.Qt.white)
        self._items = [self.pathLabel, self.numLabel, self.roiLabel, self.anLabel] + self.pluginWidgets #This list is used for changing background color and for setting all items selected.
        self.refresh()
        mdPath = os.path.join(self.Acquisition.filePath, 'AnAppPrefs.json')
        self.md: PreferencesMetadata
        try:
            self.md = PreferencesMetadata.fromJson(mdPath)
        except (JSONDecodeError, FileNotFoundError):
            self.md = PreferencesMetadata(mdPath)
        self.setInvalid(self.md.invalid)  # Update item color based on saved status. Since invalid status overrides reference status we must do this first.
        self.setReference(self.md.reference)  # We override the default automatic saving of metadata since we're just loading anyway, nothing has been changed.

    @property
    def row(self):
        """Since this can be added to a table that uses sorting we can't know that the row number will remain constant.
        This should return the correct row number."""
        return self.numLabel.row()

    def setInvalid(self, invalid: bool):
        if invalid:
            self._setItemColor(QtCore.Qt.red)
            self.md.reference = False
        else:
            self._setItemColor(QtCore.Qt.white)
        self.md.invalid = invalid

    def setReference(self, reference: bool) -> None:
        if self.isInvalid():
            return
        if reference:
            self._setItemColor(QtCore.Qt.darkGreen)
        else:
            self._setItemColor(QtCore.Qt.white)
        self.md.reference = reference

    def isInvalid(self) -> bool:
        return self.md.invalid

    def isReference(self) -> bool:
        return self.md.reference

    def setSelected(self, select: bool):
        for i in self._items:
            i.setSelected(select)

    def setHighlighted(self, select: bool):
        originalFont = self._items[0].font()
        originalFont.setBold(select)
        for i in self._items:
            i.setFont(originalFont)

    def close(self):
        try:
            self.md.close()
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.warning(f"Failed to save app metadata for {self.md.filePath}")
            logger.exception(e)

    def refresh(self):
        """Set the number of roiFile's and analyses. Update the tooltips."""
        rois = self.Acquisition.getRois()
        self.roiLabel.setNumber(len(rois))
        anNumber = 0  # This is in case the next few statements evaluate to false.
        anToolTip = ""
        if self.Acquisition.pws is not None:
            pwsAnalyses = self.Acquisition.pws.getAnalyses()
            anNumber += len(pwsAnalyses)
            if len(pwsAnalyses) != 0:
                anToolTip += "PWS:" + ', '.join(pwsAnalyses)
        if self.Acquisition.dynamics is not None:
            dynAnalyses = self.Acquisition.dynamics.getAnalyses()
            anNumber += len(dynAnalyses)
            if len(dynAnalyses) != 0:
                anToolTip += "\nDYN:" + ', '.join(dynAnalyses)
        self.anLabel.setNumber(anNumber)
        self.anLabel.setToolTip(anToolTip)
        if self.Acquisition.getNotes() != '':
            self.notesButton.setStyleSheet('QPushButton { background-color: lightgreen;}')
        else:
            self.notesButton.setStyleSheet('QPushButton { background-color: lightgrey;}')

        nameNums = [(name, num) for name, num, fformat in rois]
        if len(nameNums) > 0:
            names = set(list(zip(*nameNums))[0])
            d = {name: [num for nname, num in nameNums if nname == name] for name in names}
            self.roiLabel.setToolTip("\n".join([f'{k}: {v}' for k, v in d.items()]))

    def __del__(self):
        self.close()  # This is here just in case. realistically del rarely gets called, need to manually close each cell item.

    def _setItemColor(self, color):
        for i in self._items:
            i.setBackground(color)


class CellTableWidget(QTableWidget):
    """This is the table from which the user can select which cells to analyze, plot, etc. Each row of the table is
    represented by a CellTableWidgetItem which are stored in the self._cellItems list"""
    referencesChanged = QtCore.pyqtSignal(bool, list)
    itemsCleared = QtCore.pyqtSignal()

    def __init__(self, parent, additionalColumns: t_.Sequence[str] = None):
        super().__init__(parent)
        self.setSortingEnabled(True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._showContextMenu)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        # Columns in the form {name: (width, resizable)}
        columns = {'Path': (60, True), 'Cell#': (40, True), 'ROIs': (40, True), 'Analyses': (50, True),
                   'Notes': (40, False), 'P': (20, False), 'D': (20, False), 'F': (20, False)}
        if additionalColumns is not None:
            for colName in additionalColumns:
                columns[colName] = (50, True) # Automatically determine width from fontmetrics.
        self.setRowCount(0)
        self.setColumnCount(len(columns))
        self.setHorizontalHeaderLabels(columns.keys())
        self.verticalHeader().hide()
        [self.setColumnWidth(i, w) for i, (w, resizable) in enumerate(columns.values())]  # Set the column widths
        [self.horizontalHeader().setSectionResizeMode(i, self.horizontalHeader().Fixed) for i, (w, resizable) in enumerate(columns.values()) if not resizable] #set the notes, and p/d/f columns nonresizeable
        self.cellItems: t_.Dict[Acquisition, CellTableWidgetItem] = {}
        #This makes the items stay looking selected even when the table is inactive
        self.setStyleSheet("""QTableWidget::item:active {
                                selection-background-color: darkblue;
                                selection-color: white;}
                                
                                QTableWidget::item:inactive {
                                selection-background-color: darkblue;
                                selection-color: white;}""")
        self.palette().setColor(QPalette.Highlight, QtGui.QColor("#3a7fc2")) # This makes it so the selected cells stay colored even when the table isn't active.
        self.palette().setColor(QPalette.HighlightedText, QtCore.Qt.white)

    @property
    def selectedCellItems(self) -> t_.Tuple[CellTableWidgetItem]:
        """Returns the rows that have been selected."""
        rowIndices = [i.row() for i in self.selectedIndexes() if i.column()==0]  # This returns indexes for all items, which means we get several for a single row. Only look at the 0 column indexes.

        rowIndices.sort()
        _ = {i.row: i for i in self.cellItems.values()} #Cell items keyed by their current row position.
        return tuple(_[i] for i in rowIndices)

    def refreshCellItems(self, cells: t_.List[Acquisition] = None):
        """`Cells` indicates which cells need refreshing. If cells is None then all cells will be refreshed."""
        if cells is None:
            cells = self.cellItems.keys()
        for acq in cells:
            self.cellItems[acq].refresh()

    def addCellItems(self, items: t_.Dict[Acquisition, CellTableWidgetItem]) -> None:
        row = len(self.cellItems)
        self.setSortingEnabled(False)
        self.setRowCount(row + len(items))
        for i, item in enumerate(items.values()):
            newrow = row + i
            self.setItem(newrow, 0, item.pathLabel)
            self.setItem(newrow, 1, item.numLabel)
            self.setItem(newrow, 2, item.roiLabel)
            self.setItem(newrow, 3, item.anLabel)
            self.setCellWidget(newrow, 4, item.notesButton)
            self.setItem(newrow, 5, item.pLabel)
            self.setItem(newrow, 6, item.dLabel)
            self.setItem(newrow, 7, item.fLabel)
            for j, widg in enumerate(item.pluginWidgets):
                self.setItem(newrow, 8+j, widg)
        self.setSortingEnabled(True)
        self.cellItems.update(items)  # update the dict

    def clearCellItems(self) -> None:
        self.setRowCount(0)
        for c in self.cellItems.values():
            c.close() #This causes the cell item to save it's metadata.
        self.cellItems = {}
        self.itemsCleared.emit()

    def _showContextMenu(self, point: QtCore.QPoint):
        if len(self.selectedCellItems) > 0:
            menu = QMenu("Context Menu", parent=self)
            state = not self.selectedCellItems[0].isInvalid()
            stateString = "Disable Cell(s)" if state else "Enable Cell(s)"
            refState = not self.selectedCellItems[0].isReference()
            refStateString = "Set as Reference" if refState else "Unset as Reference"
            invalidAction = menu.addAction(stateString)
            invalidAction.triggered.connect(lambda: self._toggleSelectedCellsInvalid(state))
            refAction = menu.addAction(refStateString)
            refAction.triggered.connect(lambda: self._toggleSelectedCellsReference(refState))

            menu.addSeparator()
            mdAction = menu.addAction("Display Metadata")
            mdAction.triggered.connect(self._displayCellMetadata)
            anAction = menu.addAction("View analysis settings")
            anAction.triggered.connect(self._displayAnalysisSettings)
            if os.name == 'nt':  # Must be on windows.
                folderAction = menu.addAction("Open folder")
                folderAction.triggered.connect(lambda: [os.startfile(os.path.realpath(cellItem.Acquisition.filePath)) for cellItem in self.selectedCellItems])

            menu.addSeparator()
            delAnAction = menu.addAction("Delete analysis by name")
            delAnAction.triggered.connect(self._deleteAnalysisByName)
            delRoiAction = menu.addAction("Delete ROIs by name")
            delRoiAction.triggered.connect(self._deleteRoisByName)

            menu.exec(self.mapToGlobal(point))

    def _deleteAnalysisByName(self):
        anName, clickedOk = QInputDialog.getText(self, "Analysis Name", "Analysis name to delete")
        if not clickedOk:
            return
        deletableCells = []
        for i in self.selectedCellItems:
            if i.Acquisition.pws is not None:
                if anName in i.Acquisition.pws.getAnalyses():
                    deletableCells.append(i.Acquisition.pws)
            if i.Acquisition.dynamics is not None:
                if anName in i.Acquisition.dynamics.getAnalyses():
                    deletableCells.append(i.Acquisition.dynamics)
        if len(deletableCells)==0:
            QMessageBox.information(self, "Hmm", "No matching analysis files were found.")
        else:
            ret = ScrollableMessageBox.question(self, "Delete Analysis?",
                f"Are you sure you want to delete {anName} from:"
                f"\nPWS: {', '.join([os.path.split(i.acquisitionDirectory.filePath)[-1] for i in deletableCells if isinstance(i, ICMetaData)])}"
                f"\nDynamics: {', '.join([os.path.split(i.acquisitionDirectory.filePath)[-1] for i in deletableCells if isinstance(i, DynMetaData)])}")
            if ret == QMessageBox.Yes:
                [i.removeAnalysis(anName) for i in deletableCells]
            self.refreshCellItems()

    def _deleteRoisByName(self):
        roiName, clickeOk = QInputDialog.getText(self, "ROI Name", "ROI name to delete")
        if not clickeOk:
            return
        deletableCells = []
        for i in self.selectedCellItems:
            if roiName in [roiName for roiName, roiNum, fformat in i.Acquisition.getRois()]:
                deletableCells.append(i.Acquisition)
        if len(deletableCells)==0:
            QMessageBox.information(self, "Hmm", "No matching ROI files were found.")
        else:
            if ScrollableMessageBox.question(self, "Delete ROI?",
                                             f"Are you sure you want to delete ROI: {roiName} from: \n{', '.join([os.path.split(i.filePath)[-1] for i in deletableCells])}") == QMessageBox.Yes:
                [i.deleteRoi(roiName, roiNum) for i in deletableCells for ROIName, roiNum, fformat in i.getRois() if ROIName == roiName]
            self.refreshCellItems()

    def _displayAnalysisSettings(self):
        pwsAnalyses = set()
        for i in self.selectedCellItems:
            if i.Acquisition.pws is not None:
                #We assume that analyses with the same name have the same settings
                pwsAnalyses.update(i.Acquisition.pws.getAnalyses())
        for an in pwsAnalyses:
            for i in self.selectedCellItems:
                if i.Acquisition.pws is not None:
                    if an in i.Acquisition.pws.getAnalyses():
                        analysis: PWSAnalysisResults = i.Acquisition.pws.loadAnalysis(an)
                        settingsDict = analysis.settings.asDict()
                        settingsDict['referenceIdTag'] = analysis.referenceIdTag  # Add useful runtime information from the analysis.
                        d = DictDisplayTreeDialog(self, settingsDict, title=an)
                        d.show()
                        break
        dynAnalysis = set()
        for i in self.selectedCellItems:
            if i.Acquisition.dynamics is not None:
                dynAnalysis.update(i.Acquisition.dynamics.getAnalyses())
        for an in dynAnalysis:
            for i in self.selectedCellItems:
                if i.Acquisition.dynamics is not None:
                    if an in i.Acquisition.dynamics.getAnalyses():
                        analysis: DynamicsAnalysisResults = i.Acquisition.dynamics.loadAnalysis(an)
                        settingsDict = analysis.settings.asDict()
                        settingsDict['referenceIdTag'] = analysis.referenceIdTag  # Add useful runtime information
                        d = DictDisplayTreeDialog(self, settingsDict, title=an)
                        d.show()
                        break

    def _displayCellMetadata(self):
        for i in self.selectedCellItems:
            d = DictDisplayTreeDialog(self, i.Acquisition.pws.dict, title=os.path.join(i.path, f"Cell{i.num}"))
            d.show()

    def _toggleSelectedCellsInvalid(self, state: bool):
        changedItems = []
        for i in self.selectedCellItems:
            if i.isInvalid() != state:
                i.setInvalid(state)
                changedItems.append(i)
        if state:
            self.referencesChanged.emit(False, changedItems)

    def _toggleSelectedCellsReference(self, state: bool) -> None:
        """State indicates whether the cells are being marked as reference or as non-reference."""
        items = self.selectedCellItems
        changedItems = []
        for i in items:
            if (i.isReference() != state) and (not i.isInvalid()):
                i.setReference(state)
                changedItems.append(i)
        self.referencesChanged.emit(state, changedItems)


class ReferencesTableItem(QTableWidgetItem):
    """A single row of the reference table."""
    def __init__(self, item: CellTableWidgetItem):
        self.item = item
        super().__init__(os.path.join(item.pathLabel.text(), f'Cell{item.num}'))
        self.setToolTip(os.path.join(item.pathLabel.text(), f'Cell{item.num}'))

    def setHighlighted(self, select: bool):
        originalFont = self.font()
        originalFont.setBold(select)
        self.setFont(originalFont)


class ReferencesTable(QTableWidget):
    """This table shows all acquisitions which can be used as a reference in an analysis."""
    def __init__(self, parent: QWidget, cellTable: CellTableWidget):
        super().__init__(parent)
        #This makes the items stay looking selected even when the table is inactive
        self.setStyleSheet("""QTableWidget::item:active {   
                                selection-background-color: darkblue;
                                selection-color: white;}
                                QTableWidget::item:inactive {
                                selection-background-color: darkblue;
                                selection-color: white;}""")
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setColumnCount(1)
        self.setHorizontalHeaderLabels(('Reference',))
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.setRowCount(0)
        self.verticalHeader().hide()
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._showContextMenu)
        cellTable.referencesChanged.connect(self.updateReferences)
        cellTable.itemsCleared.connect(self._clearItems)
        self._references: t_.List[CellTableWidgetItem] = []

    def updateReferences(self, state: bool, items: t_.List[CellTableWidgetItem]):
        """state indicates if the cells are being added or being removed as references."""
        if state:
            for item in items:
                if item not in self._references:
                    row = len(self._references)
                    self.setRowCount(row + 1)
                    self.setItem(row, 0, ReferencesTableItem(item))
                    self._references.append(item)
        else:
            for item in items:
                if item in self._references:
                    self._references.remove(item)
                    item.setReference(False)
                    # find row number
                    for i in range(self.rowCount()):
                        if item is self.item(i, 0).item:
                            self.removeRow(i)
                            break

    @property
    def selectedReferenceMeta(self) -> t_.Optional[Acquisition]:
        """Returns the ICMetadata that have been selected. Return None if nothing is selected."""
        items: List[ReferencesTableItem] = self.selectedItems()
        assert len(items) <= 1
        if len(items) == 0:
            return None
        else:
            return items[0].item.Acquisition

    def _showContextMenu(self, point: QtCore.QPoint):
        items = self.selectedItems()
        if len(items) > 0:
            menu = QMenu("Context Menu", parent=self)
            refStateString = "Unset as Reference"
            refAction = menu.addAction(refStateString)
            refAction.triggered.connect(
                lambda: self.updateReferences(False, [i.item for i in self.selectedItems()]))
            menu.exec(self.mapToGlobal(point))

    def _clearItems(self):
        self.setRowCount(0)
        self._references = []




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Tuple, Dict

from PyQt5 import QtCore
from PyQt5.QtWidgets import QTableWidgetItem, QPushButton, QApplication
import matplotlib.pyplot as plt


from pwspy.analysis.compilation import (DynamicsCompilerSettings, GenericCompilerSettings, PWSCompilerSettings, AbstractCompilerSettings)
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateCompilerResults
from pwspy_gui.PWSAnalysisApp.sharedWidgets.tables import CopyableTable, NumberTableWidgetItem
from pwspy.dataTypes import Acquisition
import os


class ResultsTableItem:
    """This class embodies the results from a single ROI in Qt widget form. This is shown as a single row in the results table widget.

    Args:
        results: The object containing the results for all analysis types
        acq: A reference to the directory storing the data files.

    """
    def __init__(self, results: ConglomerateCompilerResults, acq: Acquisition):
        self.results = results
        self.acq = acq
        cellPath = os.path.split(acq.filePath)[0][len(QApplication.instance().workingDirectory) + 1:]
        cellNumber = int(acq.filePath.split('Cell')[-1])
        self.cellPathLabel = QTableWidgetItem(cellPath)
        self.cellNumLabel = NumberTableWidgetItem(cellNumber)
        # Generic results
        self.roiNameLabel = QTableWidgetItem(results.generic.roiFile.name)
        self.roiNumLabel = NumberTableWidgetItem(results.generic.roiFile.number)
        self.roiAreaLabel = NumberTableWidgetItem(results.generic.roiArea)
        # PWS related results
        pws = results.pws
        if pws is not None:
            self.pwsAnalysisNameLabel = QTableWidgetItem(pws.analysisName)
            self.rmsLabel = NumberTableWidgetItem(pws.rms)
            self.reflectanceLabel = NumberTableWidgetItem(pws.reflectance)
            self.polynomialRmsLabel = NumberTableWidgetItem(pws.polynomialRms)
            self.autoCorrelationSlopeLabel = NumberTableWidgetItem(pws.autoCorrelationSlope)
            self.rSquaredLabel = NumberTableWidgetItem(pws.rSquared)
            self.ldLabel = NumberTableWidgetItem(pws.ld)
            self.opdButton = QPushButton("OPD")
            self.opdButton.released.connect(self._plotOpd)
            if pws.opd is None:
                self.opdButton.setEnabled(False)
            self.meanSigmaRatioLabel = NumberTableWidgetItem(pws.varRatio)
        else:
            self.pwsAnalysisNameLabel = QTableWidgetItem()
            self.rmsLabel, self.reflectanceLabel, self.polynomialRmsLabel, self.autoCorrelationSlopeLabel, self.rSquaredLabel, self.ldLabel, self.meanSigmaRatioLabel = (NumberTableWidgetItem() for i in range(7))
            self.opdButton = QPushButton("OPD")
            self.opdButton.setEnabled(False)
        # Dynamics related results
        dyn = results.dyn
        if dyn is not None:
            self.dynamicsAnalysisNameLabel = QTableWidgetItem(dyn.analysisName)
            self.rms_tLabel = NumberTableWidgetItem(dyn.rms_t_squared)
            self.dynamicsReflectanceLabel = NumberTableWidgetItem(dyn.reflectance)
            self.diffusionLabel = NumberTableWidgetItem(dyn.diffusion)
        else:
            self.rms_tLabel, self.dynamicsReflectanceLabel, self.diffusionLabel = (NumberTableWidgetItem() for i in range(3))
            self.dynamicsAnalysisNameLabel = QTableWidgetItem()

    def _plotOpd(self):
        fig, ax = plt.subplots()
        ax.plot(self.results.pws.opdIndex, self.results.pws.opd)
        fig.suptitle(f"{self.cellPathLabel.text()}/Cell{self.cellNumLabel.text()}")
        fig.show()


class ResultsTable(CopyableTable):
    """
    This widget is a subclass of QTableWidget which displays all the results. It  can be copied from in CSV form.
    """
    itemsCleared = QtCore.pyqtSignal()  # this appears to be unused. Delete?

    # Columns. In the form {`name`: (`defaultVisible`, `analysisFieldName`, `compilerSettingsClass, `tooltip`)}
    columns: Dict[str, Tuple[bool, Optional[str], Optional[AbstractCompilerSettings], Optional[str]]]
    columns = {
        "Path": (False, None, None, None),
        'Cell#': (True, None, None, None),
        "PWS Analysis": (False, None, None, None),
        'ROI Name': (True, None, None, None),
        'ROI#': (True, None, None, None),
        "RMS": (True, 'rms', PWSCompilerSettings, "Primary analysis result indicating nanoscopic RI heterogeneity of sample in ROI. Defined as StdDev of the spectra"),
        'Reflectance': (True, 'reflectance', PWSCompilerSettings, "Sample reflectance averaged over the spectrum. Calculated by dividing the acquired image cube by a reference cube and then multiplying by the expected reflectance of the reference. The expected reflectance is determined by the user's choice of reference material in the analysis settings."),
        'ld': (False, 'ld', PWSCompilerSettings, "Referred to as Disorder Strength. This is proportional to RMS / AutoCorr Slope. Due to the noisiness of AutoCorr Slope this is also not very useful."),
        "AutoCorr Slope": (False, 'autoCorrelationSlope', PWSCompilerSettings, "Slope of the natural logarithm of the autocorrelation of the spectra, This is very susceptible to noise, not very useful."),
        'R^2': (False, 'rSquared', PWSCompilerSettings, "A measure the linearity of the slope of the natural logarithm of the autocorrelation function. If this is low then the AutoCorr Slope value should not be trusted."),
        'OPD': (False, 'opd', PWSCompilerSettings, "This is the Fourier transform of the spectrum. In theory this should indicate how much of the signal is contributed to by different optical path differences (OPD). Fun fact, RMS is equal to the integral of the OPD over wavenumber (k), if you are interested only in the RMS due to a specific range of OPD you can get this from summing over the appropriate range of the OPD. This is useful for removing unwanted contributions to RMS from thin films."),
        "Mean Spectra Ratio": (False, 'meanSigmaRatio', PWSCompilerSettings, "The spectral variations that we are interested in are expected to have a short spatial correlation length (neighboring pixels should not have the same spectra. However if we look at the average spectra over a cell nucleus we find that there is an overarching spectra common to the whole region. This is a measure of how much this `mean spectra` contributes to the RMS of the ROI."),
        "Poly RMS": (False, 'polynomialRms', PWSCompilerSettings, "In order to remove spectral features that are not due to interference (fluorescence, absorbance, etc.) we sometimes subtract a polynomial fit from the data before analysis. This indicates the StdDev of the polynomial fit. It's not clear how this is useful"),
        "Roi Area": (False, 'roiArea', GenericCompilerSettings, "The area of the ROI given in units of pixels. This can be converted to microns if you know the size in object space of a single pixel"),
        "Dynamics Analysis": (False, None, None, None),
        "RMS_t^2": (False, 'rms_t_squared', DynamicsCompilerSettings, "This is the primary analysis result for `Dynamics`. It is the standard deviation of the signal over time when looking at just a single wavelength."),
        "Dynamics Reflectance": (False, 'meanReflectance', DynamicsCompilerSettings, "This is the average reflectance of the ROI for the `Dynamics` measurement."),
        "Diffusion": (False, 'diffusion', DynamicsCompilerSettings, "Diffusion is calculated as the slope of the log of the autocorrelation function of a `Dynamics` measurement.")
    }

    def __init__(self):
        super().__init__()
        self.setRowCount(0)
        self.setColumnCount(len(self.columns.keys()))
        self.setHorizontalHeaderLabels(self.columns.keys())
        for i, (default, settingsName, compilerClass, tooltip) in enumerate(self.columns.values()):
            self.setColumnHidden(i, not default)
            self.horizontalHeaderItem(i).setToolTip(tooltip)
        self.verticalHeader().hide()
        self.setSortingEnabled(True)
        self._items = []

    def addItem(self, item: ResultsTableItem) -> None:
        row = len(self._items)
        self.setSortingEnabled(False)  # The fact that we are adding items assuming its the last row is a problem if sorting is on.
        self.setRowCount(row + 1)

        self.setItem(row, 0, item.cellPathLabel)
        self.setItem(row, 1, item.cellNumLabel)
        self.setItem(row, 2, item.pwsAnalysisNameLabel)
        self.setItem(row, 3, item.roiNameLabel)
        self.setItem(row, 4, item.roiNumLabel)
        self.setItem(row, 5, item.rmsLabel)
        self.setItem(row, 6, item.reflectanceLabel)
        self.setItem(row, 7, item.ldLabel)
        self.setItem(row, 8, item.autoCorrelationSlopeLabel)
        self.setItem(row, 9, item.rSquaredLabel)
        self.setCellWidget(row, 10, item.opdButton)
        self.setItem(row, 11, item.meanSigmaRatioLabel)
        self.setItem(row, 12, item.polynomialRmsLabel)
        self.setItem(row, 13, item.roiAreaLabel)
        self.setItem(row, 14, item.dynamicsAnalysisNameLabel)
        self.setItem(row, 15, item.rms_tLabel)
        self.setItem(row, 16, item.dynamicsReflectanceLabel)
        self.setItem(row, 17, item.diffusionLabel)

        self.setSortingEnabled(True)
        self._items.append(item)

    def clearCellItems(self) -> None:
        self.clearContents()
        self.setRowCount(0)
        self._items = []
        self.itemsCleared.emit()





# Copyright © 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""

@author: Nick Anthony
"""
import typing

from PyQt5 import QtCore
from PyQt5.QtCore import QModelIndex
from pwspy.utility.acquisition import SequencerStep


class TreeModel(QtCore.QAbstractItemModel):
    def __init__(self, root: SequencerStep, parent=None):
        super(TreeModel, self).__init__(parent)
        self._rootItem = SequencerStep(None, None, None)  # This will be invisible but will determine the header labels.
        self._rootItem.addChild(root)

    def invisibleRootItem(self) -> SequencerStep:
        return self._rootItem

    def columnCount(self, parent: QModelIndex) -> int:
        return 1

    def data(self, index: QModelIndex, role: int):
        if not index.isValid():
            return None
        if role != QtCore.Qt.DisplayRole:  # We only support this role type. Return the SequencerStep itself as the data.
            return None
        item: SequencerStep = index.internalPointer()
        return item

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable

    def setData(self, index: QModelIndex, value: typing.Any, role: int = ...) -> bool:
        return True  # We don't allow setting data. Always report success.

    def headerData(self, section, orientation, role):
        return "Steps"
        # if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
        #     return self._rootItem.data(section)
        # return None

    def index(self, row: int, column: int, parent: QModelIndex):
        # if not self.hasIndex(row, column, parent): #This was causing bugs
        #     return QtCore.QModelIndex()
        if parent.isValid():
            parentItem = parent.internalPointer()
        else:
            parentItem = self._rootItem
        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QtCore.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()
        childItem = index.internalPointer()
        parentItem = childItem.parent()
        if parentItem == self._rootItem:
            return QtCore.QModelIndex()
        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent: QModelIndex):
        if parent.column() > 0:
            return 0
        if not parent.isValid():
            return self._rootItem.childCount()
        else:
            return parent.internalPointer().childCount()




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import shutil
from glob import glob
from PyQt5.QtWidgets import QApplication, QMessageBox
from PyQt5 import QtCore
from pwspy_gui.PWSAnalysisApp.App import PWSApp
from pwspy_gui.PWSAnalysisApp import applicationVars
from datetime import datetime
from pwspy_gui.PWSAnalysisApp import resources
from pwspy.analysis import defaultSettingsPath


def _setupDataDirectories():
    if not os.path.exists(applicationVars.dataDirectory):
        os.mkdir(applicationVars.dataDirectory)
    if not os.path.exists(applicationVars.analysisSettingsDirectory):
        os.mkdir(applicationVars.analysisSettingsDirectory)
    settingsFiles = glob(os.path.join(defaultSettingsPath, '*.json'))
    if len(settingsFiles) == 0:
        raise Exception("Warning: Could not find any analysis settings presets.")
    for f in settingsFiles:  # This will overwrite any existing preset files in the application directory with the defaults in the source code.
        shutil.copyfile(f, os.path.join(applicationVars.analysisSettingsDirectory, os.path.split(f)[-1]))
    if not os.path.exists(applicationVars.extraReflectionDirectory):
        os.mkdir(applicationVars.extraReflectionDirectory)
        with open(os.path.join(applicationVars.extraReflectionDirectory, 'readme.txt'), 'w') as f:
            f.write("""Extra reflection `data cubes` and an index file are stored on the Backman Lab google drive account.
            Download the index file and any data cube you plan to use to this folder.""")
    if not os.path.exists(applicationVars.googleDriveAuthPath):
        os.mkdir(applicationVars.googleDriveAuthPath)
        shutil.copyfile(os.path.join(resources, 'credentials.json'),
                        os.path.join(applicationVars.googleDriveAuthPath, 'credentials.json'))


def main():
    import sys
    import getopt

    debugMode = False
    try:
        opts, args = getopt.getopt(sys.argv[1:], '', ['debug'])
        if len(opts) > 0:  # If any arguments were found
            optNames, optVals = zip(*opts)
            if '--debug' in optNames:
                debugMode = True
    except getopt.GetoptError as e:
        print(e)
        sys.exit(1)

    def isIpython():
        try:
            return __IPYTHON__  # Only defined if we are running within ipython
        except NameError:
            return False

    _setupDataDirectories()  # this must happen before the logger can be instantiated
    logger = logging.getLogger()  # We use the root logger so that all loggers in pwspy will be captured.

    # This prevents errors from happening silently. Found on stack overflow.
    sys.excepthook_backup = sys.excepthook
    def exception_hook(exctype, value, traceBack):
        logger.exception("Unhandled Exception! :", exc_info=value, stack_info=True)
        sys.excepthook_backup(exctype, value, traceBack)  # Run the error through the default exception hook
        sys.exit(1)
    sys.excepthook = exception_hook

    logger.addHandler(logging.StreamHandler(sys.stdout))
    fHandler = logging.FileHandler(os.path.join(applicationVars.dataDirectory, f'log{datetime.now().strftime("%d%m%Y%H%M%S")}.txt'))
    fHandler.setFormatter(logging.Formatter('%(levelname)s: %(asctime)s %(name)s.%(funcName)s(%(lineno)d) - %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
    logger.addHandler(fHandler)
    logger.setLevel(logging.INFO)
    if debugMode:
        logging.getLogger("pwspy_gui").setLevel(logging.DEBUG)
        logging.getLogger("pwspy").setLevel(logging.DEBUG)
        logger.info("Logger set to debug mode.")

    try:
        os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"  # TODO replace these options with proper high dpi handling. no pixel specific widths.
        QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
        logger.debug("About to construct `PWSApp`")
        app = PWSApp(sys.argv)
        logger.debug("Finished constructing `PWSApp`")

        #Testing script
        # app.changeDirectory(r'\\backmanlabnas.myqnapcloud.com\home\Year3\zstack_focusSensitivity\again', False)
        # app.setSelectedCells(app.getLoadedCells()[:3])
        # app.plotSelectedCells()
        # app.window.plots._startRoiDrawing()

        if not isIpython():  # IPython runs its own QApplication so we handle things slightly different.
            sys.exit(app.exec_())
    except Exception as e:  # Save error to text file.
        logger.exception(e)
        msg = f"Error Occurred. Please check the log in: {applicationVars.dataDirectory}"
        msgBox = QMessageBox.information(None, 'Crash!', msg)
        raise e


if __name__ == '__main__':
    main()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
from datetime import datetime

from pwspy_gui.ExtraReflectanceCreator.app import ERApp
from pwspy_gui import appPath


def isIpython():
    try:
        return __IPYTHON__
    except:
        return False


def main():
    import sys

    # This prevents errors from happening silently.
    sys.excepthook_backup = sys.excepthook
    def exception_hook(exctype, value, tb):
        print(exctype, value, tb)
        sys.excepthook_backup(exctype, value, tb)
        sys.exit(1)
    sys.excepthook = exception_hook

    logger = logging.getLogger()  # We use the root logger so that all loggers in pwspy will be captured.
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    appDataPath = os.path.join(appPath, 'ExtraReflectanceCreatorData')
    if not os.path.exists(appDataPath):
        os.mkdir(appDataPath)
    fHandler = logging.FileHandler(os.path.join(appDataPath, f'log{datetime.now().strftime("%d%m%Y%H%M%S")}.txt'))
    fHandler.setFormatter(logging.Formatter('%(levelname)s: %(asctime)s %(name)s.%(funcName)s(%(lineno)d) - %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
    logger.addHandler(fHandler)

    if isIpython():
        app = ERApp(sys.argv)
    else:
        app = ERApp(sys.argv)
        sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import json
import shutil

from PyQt5 import QtGui
from PyQt5.QtWidgets import QDialog, QFileDialog, QWidget, QLineEdit, QPushButton, QLabel, QGridLayout, QMessageBox

from pwspy_gui import resources
from pwspy.dataTypes import ICMetaData
from typing import List
import os
import random


class Blinder:
    """A class that, given a list of ICMetadata and the root directory that their files are under, will create randomly
    numbered symlinks in `outDir` and an index that can be used to trace back to the original. Useful for creating
    blinded experiments."""
    def __init__(self, cells: List[ICMetaData], homeDir: str, outDir: str):
        indexPath = os.path.join(homeDir, 'blindedIndex.json')
        if os.path.exists(indexPath):
            raise ValueError(f"A `blindedIndex.json` file already exists in {homeDir}.")
        if not os.path.exists(outDir):
            os.mkdir(outDir)
        try:
            paths = [c.filePath for c in cells]
            nums = list(range(1, len(paths)+1))
            assert len(nums) == len(paths)
            random.shuffle(nums)  # nums is now a random list of non-repeating numbers from one up to the number of cells.
            d = {'outputDirectory': outDir, 'index': []}
            for path, num in zip(paths, nums):
                newPath = os.path.join(outDir, f'Cell{num}')
                os.symlink(path, newPath)
                d['index'].append({path: f'Cell{num}'})
        except Exception as e:
            shutil.rmtree(outDir)
            raise e
        with open(indexPath, 'w') as f:
            json.dump(d, f)


class BlinderDialog(QDialog):
    """This dialog asks the user for the information that is needed in order to perform a blinding with the `Blinder`
    class."""
    def __init__(self, parent: QWidget, homeDir: str, cells: List[ICMetaData]):
        self.parent = parent
        self.homeDir = homeDir
        self.cells = cells
        super().__init__(self.parent)
        self.setModal(True)
        self.pathEdit = QLineEdit(self)
        self.browseButton = QPushButton(QtGui.QIcon(os.path.join(resources, 'folder.svg')), '')
        self.browseButton.released.connect(self._getPath)
        self.okButton = QPushButton('Ok', self)
        self.okButton.released.connect(self.accept)
        self.okButton.setMaximumWidth(100)
        layout = QGridLayout(self)
        layout.addWidget(self.browseButton, 0, 0)
        layout.addWidget(self.pathEdit, 0, 1)
        layout.addWidget(self.okButton, 1, 1, 1, 2)

        newDir = self._getPath()
        self.pathEdit.setText(newDir)

    def _getPath(self) -> str:
        newDir = QFileDialog.getExistingDirectory(self.parent, "Select location for new blinded directory", self.pathEdit.text())
        return newDir

    def accept(self) -> None:
        try:
            outDir = self.pathEdit.text()
            b = Blinder(self.cells, self.homeDir, outDir)
            super().accept()
        except Exception as e:
            msg = QMessageBox.warning(self, 'Uh Oh', str(e))



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Tue Nov 20 13:13:34 2018

@author: Nick Anthony

This functionality has now been replaced in the ERCreator application. This is just an example script.
This script is now outdated and will not work. See the `generateExtraReflectance` example for a more up to date example.
"""


from pwspy.dataTypes import CameraCorrection
from pwspy.utility.fileIO import loadAndProcess
from pwspy.utility.reflection import reflectanceHelper, Material
import pwspy.utility.reflection.extraReflectance as er
from glob import glob
import matplotlib.pyplot as plt
import os
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd
import random


def processIm(im):
    im.correctCameraEffects(CameraCorrection(100))
#    im.subtractDarkCounts(2000)
#    im.data = np.polynomial.polynomial.polyval(im.data, [0, 0.977241216, 1.73E-06, 1.70E-11]) #LCPWS 1 linearization
    im.normalizeByExposure()
    im.filterDust(6)
    return im


if __name__ == '__main__':
    __spec__ = None
    plt.interactive(True)
    
    exclude = [(Material.Water, Material.Ipa), (Material.Water, Material.Ethanol)]
    rootDir = r'/home/nick/Desktop/4matExtraReflection/LCPWS1/1_23_2019'
    produceRextraCube = True
    plotResults = True
    
    materials = [('air', Material.Air), ('water', Material.Water), ('ipa', Material.Ipa), ('ethanol', Material.Ethanol)]
    
    fig, axes = plt.subplots(ncols=2)
    axes[0].set_ylabel('n')
    axes[0].set_xlabel('nm')
    axes[0].set_title("Index of Refraction")
    axes[1].set_ylabel('reflectance')
    axes[1].set_xlabel('nm')
    axes[1].set_title("Glass Interface Reflectance")
    [axes[0].plot(reflectanceHelper.getRefractiveIndex(mat), label=matName) for matName, mat in materials]
    axes[0].legend()
    [axes[1].plot(reflectanceHelper.getReflectance(mat, Material.Glass).index, reflectanceHelper.getReflectance(mat, Material.Glass), label=matName) for matName, mat in materials]
    axes[1].legend()
    
    fileFrame = pd.DataFrame([{'setting': 'none', 'material': m, 'cube': cube} for m in materials for cube in glob(os.path.join(rootDir, m, 'Cell*'))])
    cubes = loadAndProcess(fileFrame, processIm, parallel=True)

    theoryR = er.getTheoreticalReflectances(list(zip(*materials))[1], cubes['cube'][0].wavelengths, 0.52)
    matCombos = er.generateMaterialCombos(list(zip(*materials))[1], excludedCombos=exclude)
    if plotResults:
        mask = random.choice(cubes['cube']).selectLassoRoi()
        er.plotExtraReflection(cubes, theoryR, matCombos, 0.52, mask, plotReflectionImages=False)
        with PdfPages(os.path.join(rootDir, "figs.pdf")) as pp:
            for i in plt.get_fignums():
                f = plt.figure(i)
                f.set_size_inches(9, 9)
                pp.savefig(f)
    if produceRextraCube:
        for sett in set(cubes['setting']):
            allCombos = er.getAllCubeCombos(matCombos, cubes[cubes['setting'] == sett])
            erCube, rextras, plots = er.generateRExtraCubes(allCombos, theoryR, 0.52)
            erCube.toHdfFile(rootDir, f'rextra_{sett}')



# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import hashlib
import os
from datetime import datetime
from glob import glob
import json
import typing as t_
from PyQt5.QtWidgets import QWidget
from matplotlib import animation

from pwspy.dataTypes import CameraCorrection, Acquisition, ICMetaData, PwsCube
from pwspy_gui.ExtraReflectanceCreator.widgets.dialog import IndexInfoForm
from pwspy.dataTypes import Roi
from pwspy import dateTimeFormat
from pwspy.utility.reflection import Material
from pwspy.utility.fileIO import loadAndProcess
import pwspy.utility.reflection.extraReflectance as er
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import logging
from mpl_qt_viz.visualizers import PlotNd, DockablePlotWindow
import pathlib as pl


class Directory:
    def __init__(self, df: pd.DataFrame, camCorr: CameraCorrection):
        self.dataframe = df
        self.cameraCorrection = camCorr


def scanDirectory(directory: str) -> Directory:
    """
    Scan a folder for data in the format expected by this application. We expect multiple layers of subfolders.
    Layer 1: Folders named by date.
    Layer 2: Folders named by material imaged. Must be supported by pwspy.utility.reflectance.reflectanceHelper and in the `matMap` variable.
    Layer 3: "Cell{x}" folders each containing a single acquisition.

    Args:
        directory: The file path to be scanned.

    Returns:
        A `Directory` object for the scanned directory.
    """
    try:
        cam = CameraCorrection.fromJsonFile(os.path.join(directory, 'cameraCorrection.json'))
    except Exception as e:
        logger = logging.getLogger(__name__)
        logger.exception(e)
        raise Exception(f"Could not load a camera correction at {directory}. Please add a `cameraCorrection.json` file to describe how to correct for dark counts and nonlinearity.")
    files = glob(os.path.join(directory, '*', '*', 'Cell*'))
    rows = []
    matMap = {'air': Material.Air, 'water': Material.Water, 'ipa': Material.Ipa, 'ethanol': Material.Ethanol, 'glass': Material.Glass,
              'methanol': Material.Methanol}
    for file in files:
        filelist = pl.Path(file).parts  # Split file into components.
        s = filelist[-3]
        m = matMap[filelist[-2]]
        file = Acquisition(file).pws.filePath  # old pws is saved directly in the "Cell{X}" folder. new pws is saved in "Cell{x}/PWS" the Acquisition class helps us abstract that out and be compatible with both.
        rows.append({'setting': s, 'material': m, 'cube': file})
    df = pd.DataFrame(rows)
    return Directory(df, cam)


class DataProvider:
    def __init__(self, df: pd.DataFrame, camCorr: CameraCorrection):
        self._df = df
        self._cameraCorrection = camCorr
        self._cubes = None

    def getCubes(self):
        return self._cubes

    def getDataFrame(self):
        return self._df

    def loadCubes(self, includeSettings: t_.List[str], binning: int, parallelProcessing: bool):
        df = self._df[self._df['setting'].isin(includeSettings)]
        if binning is None:
            args = {'correction': None, 'binning': None}
            for cube in df['cube']:
                md = ICMetaData.loadAny(cube)
                if md.binning is None:
                    raise Exception("No binning metadata found. Please specify a binning setting.")
                elif md.cameraCorrection is None:
                    raise Exception(
                        "No camera correction metadata found. Please specify a binning setting, in this case the application will use the camera correction stored in the cameraCorrection.json file of the calibration folder")
        else:
            args = {'correction': self._cameraCorrection, 'binning': binning}
        self._cubes = loadAndProcess(df, self._processIm, parallel=parallelProcessing, procArgs=[args])
        self._cubes['material'] = self._cubes['material'].astype('category')

    def _processIm(self, im: PwsCube, kwargs) -> PwsCube:
        """
        This processor function may be run in parallel to pre-process each raw image.

        Args:
            im: The `PwsCube` to be preprocessed.
            kwargs: These keyword arguments are passed to `PwsCube.correctCameraEffects`

        Returns:
            The same PwsCube that was provided as input.
        """
        im.correctCameraEffects(**kwargs)
        im.normalizeByExposure()
        try:
            im.filterDust(0.8)  # in microns
        except ValueError:
            logger = logging.getLogger(__name__)
            logger.warning("No pixel size metadata found. assuming a gaussian filter radius of 6 pixels = 1 sigma.")
            im.filterDust(6, pixelSize=1)  # Do the filtering in units of pixels if no auto pixelsize was found
        return im


class ERWorkFlow:
    """
    This class serves as an adapter between the complicated operations available in the pwspy.utility.relfection.extraReflectance
    module and the UI of the ERCreator app.

    Args:
        workingDir: The folder to scan for images. Each subfolder of workingDir should be named by the corresponding system / configuration.
            The contents of each subfolder should align with the description in `scanDirectory`
        homeDir: TODO
    """
    def __init__(self, workingDir: str, homeDir: str):
        self.fileStruct = self.df = self.cameraCorrection = self.currDir = self.plotnds = self.anims = None
        self.figs = []
        self.dataprovider = None
        self.homeDir = homeDir
        folders = [i for i in glob(os.path.join(workingDir, '*')) if os.path.isdir(i)]
        settings = [os.path.split(i)[-1] for i in folders]
        self.fileStruct = {}
        for f, s in zip(folders, settings):
            self.fileStruct[s] = scanDirectory(f)

    def loadIfNeeded(self, includeSettings: t_.List[str], binning: int, parallelProcessing: bool):
        if self.dataprovider.getCubes() is None:
            self.loadCubes(includeSettings, binning, parallelProcessing)

    def invalidateCubes(self):
        """Clear the cached data requiring that data is reloaded from file."""
        self.dataprovider._cubes = None

    def deleteFigures(self):
        for fig in self.figs:
            if isinstance(fig, plt.Figure):
                plt.close(fig)
            elif isinstance(fig, QWidget):
                fig.close()
            else:
                raise TypeError(f"Type {type(fig)} shouldn't be here, what's going on?")
        self.figs = []

    def loadCubes(self, includeSettings: t_.List[str], binning: int, parallelProcessing: bool):
        self.dataprovider.loadCubes(includeSettings, binning, parallelProcessing)

    def plot(self, includeSettings: t_.List[str], binning: int, parallelProcessing: bool, numericalAperture: float, saveToPdf: bool = False, saveDir: str = None):
        self.loadIfNeeded(includeSettings, binning, parallelProcessing)
        cubes = self.dataprovider.getCubes()
        materials = set(cubes['material'])
        theoryR = er.getTheoreticalReflectances(materials,
                                                cubes['cube'].iloc[0].wavelengths, numericalAperture)  # Theoretical reflectances
        matCombos = er.generateMaterialCombos(materials)

        print("Select an ROI")
        roi = cubes['cube'].sample(n=1).iloc[0].selectLassoROI()  # Select an ROI to analyze
        cubeDict = cubes.groupby('setting').apply(lambda df: df.groupby('material')['cube'].apply(list).to_dict()).to_dict()  # Transform data frame to a dict of dicts of lists for input to `plot`
        self.figs.extend(er.plotExtraReflection(cubeDict, theoryR, matCombos, numericalAperture, roi))
        if saveToPdf:
            with PdfPages(os.path.join(saveDir, f"fig_{datetime.strftime(datetime.now(), '%d-%m-%Y %HH%MM%SS')}.pdf")) as pp:
                for i in plt.get_fignums():
                    f = plt.figure(i)
                    f.set_size_inches(9, 9)
                    pp.savefig(f)

    def save(self, includeSettings: t_.List[str], binning: int, parallelProcessing: bool, numericalAperture: float, parentWidget: QWidget):
        self.loadIfNeeded(includeSettings, binning, parallelProcessing)
        cubes = self.dataprovider.getCubes()
        settings = set(cubes['setting'])
        for setting in settings:
            sCubes = cubes[cubes['setting'] == setting]
            materials = set(sCubes['material'])
            theoryR = er.getTheoreticalReflectances(materials,
                                                    sCubes['cube'].iloc[0].wavelengths, numericalAperture)  # Theoretical reflectances
            matCombos = er.generateMaterialCombos(materials)
            matCubes = sCubes.groupby('material')['cube'].apply(list).to_dict()
            combos = er.getAllCubeCombos(matCombos, matCubes)
            erCube, rExtraDict = er.generateRExtraCubes(combos, theoryR, numericalAperture)
            dock = DockablePlotWindow(title=setting)
            dock.addWidget(
                PlotNd(erCube.data, title='Mean',
                       indices=[range(erCube.data.shape[0]), range(erCube.data.shape[1]),
                                erCube.wavelengths]),
                title='Mean'
            )
            for matCombo, rExtraArr in rExtraDict.items():
                dock.addWidget(
                    PlotNd(rExtraArr, title=matCombo,
                           indices=[range(erCube.data.shape[0]), range(erCube.data.shape[1]),
                                    erCube.wavelengths]),
                    title=str(matCombo)
                )
            logger = logging.getLogger(__name__)
            logger.info(f"Final data max is {erCube.data.max()}")
            logger.info(f"Final data min is {erCube.data.min()}")
            self.figs.append(dock)  # keep track of opened figures.
            saveName = f'{self.currDir}-{setting}'
            dialog = IndexInfoForm(f'{self.currDir}-{setting}', erCube.metadata.idTag, parent=parentWidget)

            def accepted(dialog, erCube, saveName):
                erCube.metadata.inheritedMetadata['description'] = dialog.description
                erCube.toHdfFile(self.homeDir, saveName)
                self.updateIndex(saveName, erCube.metadata.idTag, dialog.description, erCube.metadata.dirName2Directory('', saveName))

            dialog.accepted.connect(lambda dlg=dialog, ercube=erCube, savename=saveName: accepted(dlg, ercube, savename))
            dialog.show()

    def updateIndex(self, saveName: str, idTag: str, description: str, filePath: str):
        with open(os.path.join(self.homeDir, 'index.json'), 'r') as f:
            index = json.load(f)
        md5hash = hashlib.md5()
        with open(os.path.join(self.homeDir, filePath), 'rb') as f:
            md5hash.update(f.read())
        md5 = md5hash.hexdigest() #The md5 checksum as a string of hex.
        cubes = index['reflectanceCubes']
        newEntry = {'fileName': filePath,
                    'description': description,
                    'idTag': idTag,
                    'name': saveName,
                    'md5': md5}
        cubes.append(newEntry)
        index['reflectanceCubes'] = cubes
        index['creationDate'] = datetime.strftime(datetime.now(), dateTimeFormat)
        with open(os.path.join(self.homeDir, 'index.json'), 'w') as f:
            json.dump(index, f, indent=4)

    def compareDates(self, includeSettings: t_.List[str], binning: int, parallelProcessing: bool):
        self.loadIfNeeded(includeSettings, binning, parallelProcessing)
        anis = []
        figs = []  # These lists just maintain references to matplotlib figures to keep them responsive.
        cubes = self.dataprovider.getCubes()
        verts = cubes['cube'].sample(n=1).iloc[0].selectLassoROI()  # Select a random of the selected cubes and use it to prompt the user for an analysis ROI
        mask = Roi.fromVerts(verts=verts, dataShape=cubes['cube'].sample(n=1).iloc[0].data.shape[:-1])
        for mat in set(cubes['material']):
            c = cubes[cubes['material'] == mat]
            fig, ax = plt.subplots()
            fig.suptitle(mat.name)
            ax.set_xlabel("Wavelength (nm)")
            ax.set_ylabel("Counts/ms")
            fig2, ax2 = plt.subplots()
            fig2.suptitle(mat.name)
            figs.extend([fig, fig2])
            anims = []
            for i, row in c.iterrows():
                im = row['cube']
                spectra = im.getMeanSpectra(mask)[0]
                ax.plot(im.wavelengths, spectra, label=row['setting'])
                anims.append((ax2.imshow(im.data.mean(axis=2), animated=True,
                                         clim=[np.percentile(im.data, .5), np.percentile(im.data, 99.5)]),
                              ax2.text(40, 40, row['setting'])))
            ax.legend()
            anis.append(animation.ArtistAnimation(fig2, anims, interval=1000, blit=False))
        self.anims = anis

        self.figs.extend(figs) #Keep track of opened figures.

    def directoryChanged(self, directory: str) -> t_.Set[str]:
        """
        Args:
            directory:

        Returns:
            A list of the `settings` found in the new directory.
        """
        self.currDir = directory
        directory = self.fileStruct[directory]
        self.dataprovider = DataProvider(directory.dataframe, directory.cameraCorrection)
        return set(self.dataprovider.getDataFrame()['setting'])



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Optional

import pandas
import typing
if typing.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager, ERDownloader
from pwspy_gui.sharedWidgets.extraReflectionManager._ERDataDirectory import ERDataDirectory, EROnlineDirectory
from enum import Enum


class ERDataComparator:
    """A class to compare the local directory to the online directory.
    Args:
        downloader (Optional[ERDownloader]): Handles communication with GoogleDrive, if operating in offline mode this should be None
        directory (str): The file path where the files are stored locally.
    """

    class ComparisonStatus(Enum):
        LocalOnly = "Local Only"
        OnlineOnly = "Online Only"
        Md5Mismatch = 'MD5 Mismatch'
        Match = "Match"  # This is what we hope to see.

    def __init__(self, downloader: Optional[ERDownloader], directory: str):
        self.local: ERDataDirectory = ERDataDirectory(directory)
        self.online: Optional[EROnlineDirectory] = None if downloader is None else EROnlineDirectory(downloader)

    def updateIndexes(self):
        self.local.updateIndex()
        if self.online is not None:
            self.online.updateIndex()

    def compare(self) -> pandas.DataFrame:
        """Scans local and online files to put together an idea of the status."""
        localStatus = self.local.getFileStatus()
        if self.online is not None:
            onlineStatus = self.online.getFileStatus()
            status = pandas.merge(localStatus, onlineStatus, how='outer', on='idTag')
            status['Index Comparison'] = status.apply(lambda row: self._dataFrameCompare(row), axis=1)
            status = status[['idTag', 'Local Status', 'Online Status', 'Index Comparison']]  # Set the column order
        else:
            status = localStatus
            status['Index Comparison'] = self.ComparisonStatus.LocalOnly.value
            status = status[['idTag', 'Local Status', 'Index Comparison']]  # Set the column order
        return status

    def _dataFrameCompare(self, row) -> str:
        try:
            self.local.index.getItemFromIdTag(row['idTag'])
        except:
            return self.ComparisonStatus.OnlineOnly.value
        try:
            self.online.index.getItemFromIdTag(row['idTag'])
        except:
            return self.ComparisonStatus.LocalOnly.value
        if self.local.index.getItemFromIdTag(row['idTag']).md5 != self.online.index.getItemFromIdTag(row['idTag']).md5:
            return self.ComparisonStatus.Md5Mismatch.value
        else:
            return self.ComparisonStatus.Match.value



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import typing
from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QScrollArea, QGridLayout, QLineEdit, QLabel, QGroupBox, QHBoxLayout, QCheckBox, QMessageBox

import pwspy.analysis.dynamics
from pwspy_gui.PWSAnalysisApp._dockWidgets import CellSelectorDock
from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector
from pwspy.utility.reflection import Material
from ._AbstractSettingsFrame import AbstractSettingsFrame

from ._sharedWidgets import ExtraReflectanceSelector, VerticallyCompressedWidget, HardwareCorrections
from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.runtimeSettings import DynamicsRuntimeAnalysisSettings
if typing.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager


class DynamicsSettingsFrame(QScrollArea, AbstractSettingsFrame):
    def __init__(self, erManager: ERManager, cellSelector: CellSelector):
        super().__init__()
        self.cellSelector = cellSelector

        self._frame = VerticallyCompressedWidget(self)
        self._layout = QGridLayout()
        self._frame.setLayout(self._layout)
        self._frame.setFixedWidth(350)
        self.setMinimumWidth(self._frame.width()+5)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setWidget(self._frame)

        row = 0
        self._analysisNameEdit = QLineEdit()
        self._layout.addWidget(QLabel("Analysis Name: "), row, 0, 1, 1)
        self._layout.addWidget(self._analysisNameEdit, row, 1, 1, 1)
        row += 1

        self.hardwareCorrections = HardwareCorrections(self)
        self.hardwareCorrections.stateChanged.connect(self._updateSize)
        self._layout.addWidget(self.hardwareCorrections, row, 0, 1, 4)
        row += 1

        self.extraReflection = ExtraReflectanceSelector(self, erManager)
        self.extraReflection.loadFromSettings(0.52, Material.Water, None)
        self._layout.addWidget(self.extraReflection, row, 0, 1, 4)
        row += 1

        '''Use relative or absolute units of reflectance'''
        self.scaling = QGroupBox("Scaling")
        self.scaling.setLayout(QHBoxLayout())
        self.scaling.layout().setContentsMargins(2, 2, 2, 5)
        self.relativeUnits = QCheckBox("Use Relative Units", self)
        self.relativeUnits.setToolTip("If checked then reflectance (and therefore all other parameters) will be scaled such that any reflectance matching that\n"
                                      "of the reference image will be 1. If left unchecked then the `Reference Material` will be used to scale reflectance to\n"
                                      "match the actual physical reflectance of the sample.")
        self.relativeUnits.setChecked(True)
        self.scaling.layout().addWidget(self.relativeUnits)
        self._layout.addWidget(self.scaling, row, 0, 1, 4)

        self._updateSize()

    def showEvent(self, a0: QtGui.QShowEvent) -> None:
        super().showEvent(a0)
        self._updateSize() #For some reason this must be done here and in the __init__ for it to start up properly.

    def _updateSize(self):
        height = 100  # give this much excess room.
        height += self.hardwareCorrections.height()
        height += self.extraReflection.height()
        height += self.scaling.height()
        self._frame.setFixedHeight(height)

    def loadFromSettings(self, settings: pwspy.analysis.dynamics.DynamicsAnalysisSettings):
        self.extraReflection.loadFromSettings(settings.numericalAperture, settings.referenceMaterial, settings.extraReflectanceId)
        self.relativeUnits.setCheckState(2 if settings.relativeUnits else 0)
        self.hardwareCorrections.loadCameraCorrection(settings.cameraCorrection)

    def getSettings(self) -> DynamicsRuntimeAnalysisSettings:
        erMetadata, refMaterial, numericalAperture = self.extraReflection.getSettings()
        refMeta = self.cellSelector.getSelectedReferenceMeta()
        cellMeta = self.cellSelector.getSelectedCellMetas()
        name = self._analysisNameEdit.text()
        if refMeta is None:
            raise ValueError('Please select a reference Cell.')
        if name == '':
            raise ValueError("Please give your analysis a name.")
        if len(cellMeta) == 0:
            raise ValueError('Please select cells to analyse.')
        refMeta = refMeta.dynamics
        cellMeta = [i.dynamics for i in cellMeta if i.dynamics is not None]  # If we select some acquisitions that don't have dynamics then they'll get stripped out here
        if refMeta is None:
            raise ValueError("The selected reference acquisition has no valid Dynamics data.")
        if len(cellMeta) == 0:
            raise ValueError("No valid Dynamics acquisitions were selected.")
        return DynamicsRuntimeAnalysisSettings(settings=pwspy.analysis.dynamics.DynamicsAnalysisSettings(extraReflectanceId=erMetadata.idTag if erMetadata is not None else None,
                                                                                                                                 referenceMaterial=refMaterial,
                                                                                                                                 numericalAperture=numericalAperture,
                                                                                                                                 relativeUnits=self.relativeUnits.checkState() != 0,
                                                                                                                                 cameraCorrection=self.hardwareCorrections.getCameraCorrection()),
                                                                       extraReflectanceMetadata=erMetadata,
                                                                       referenceMetadata=refMeta,
                                                                       cellMetadata=cellMeta,
                                                                       analysisName=name)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import logging
import pickle
import re
import typing
from dataclasses import dataclass
from PyQt5.QtCore import pyqtSignal, Qt, QMimeData
from shapely.geometry import Polygon as shapelyPolygon
from matplotlib.backend_bases import KeyEvent, MouseEvent
from matplotlib.patches import PathPatch
import numpy as np
from PyQt5.QtGui import QCursor, QValidator
from PyQt5.QtWidgets import QMenu, QAction, QComboBox, QLabel, QPushButton, QHBoxLayout, QWidget, QVBoxLayout, QApplication, QMessageBox
from PyQt5 import QtCore
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._bigPlot import BigPlot
from mpl_qt_viz.roiSelection import PolygonModifier, MovingModifier
import pwspy.dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp._roiManager import _DefaultROIManager, ROIManager
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._sinCityExporter import SinCityDlg
import descartes
import os


@dataclass
class RoiParams:
    roiFile: pwsdt.RoiFile
    polygon: PathPatch
    selected: bool


class RoiPlot(QWidget):
    """Adds GUI handling for ROIs."""
    roiDeleted = pyqtSignal(pwsdt.Acquisition, pwsdt.RoiFile)  # Indicates that an ROI deletion was initiated by this widget.
    roiModified = pyqtSignal(pwsdt.Acquisition, pwsdt.RoiFile)  # Indicates that an ROI modification was initiated by this widget.
    roiCreated = pyqtSignal(pwsdt.Acquisition, pwsdt.RoiFile)  # Indicates that an ROI modification was created by this widget.

    def __init__(self, Acquisition: pwsdt.Acquisition, data: np.ndarray, roiManager: ROIManager, parent=None, flags: QtCore.Qt.WindowFlags = None):
        if flags is not None:
            super().__init__(parent, flags=flags)
        else:
            super().__init__(parent=parent)
        self._plotWidget = BigPlot(data, self)
        self.im = self._plotWidget.im
        self.ax = self._plotWidget.ax

        self.rois: typing.List[RoiParams] = []  # This list holds information about the ROIs that are currently displayed.

        self.roiFilter = QComboBox(self)
        self.roiFilter.setEditable(True)
        self.roiFilter.setValidator(WhiteSpaceValidator())

        self.exportButton = QPushButton("Export")
        self.exportButton.released.connect(self._exportAction)

        layout = QVBoxLayout()
        l = QHBoxLayout()
        l.addWidget(QLabel("Roi"), alignment=QtCore.Qt.AlignRight)
        l.addWidget(self.roiFilter)
        l.addWidget(self.exportButton)
        layout.addLayout(l)
        layout.addWidget(self._plotWidget)
        self.setLayout(layout)

        self.metadata: pwsdt.Acquisition = None
        self.setMetadata(Acquisition)

        self.annot = self._plotWidget.ax.annotate("", xy=(0, 0), xytext=(20, 20), textcoords="offset points",
                            bbox=dict(boxstyle="round", fc="w"),
                            arrowprops=dict(arrowstyle="->"))

        self._toggleCids = None
        self.enableHoverAnnotation(True)

        self._roiManager = roiManager
        self._roiManager.roiRemoved.connect(self._onRoiRemoved)
        self._roiManager.roiUpdated.connect(self._onRoiUpdated)
        self._roiManager.roiCreated.connect(self._onRoiCreated)

    def getImageData(self) -> np.ndarray:
        return self._plotWidget.data

    def setImageData(self, data: np.ndarray):
        self._plotWidget.setImageData(data)

    def setMetadata(self, metadata: pwsdt.Acquisition):
        """Refresh the ROIs based on a new metadata. Also needs to be provided with the data for the image to display."""
        self.metadata = metadata
        self._clearRois()
        currentSel = self.roiFilter.currentText()
        # updateFilter
        try:
            self.roiFilter.currentIndexChanged.disconnect()  # Without this line the roiFilter.clear() line is very slow.
        except Exception:
            pass  # if the signal hasn't yet been connected we'll get an error. ignore it.
        self.roiFilter.clear()
        self.roiFilter.addItem(' ')
        self.roiFilter.addItem('.*')
        rois = self.metadata.getRois()
        roiNames = set(list(zip(*rois))[0]) if len(rois) > 0 else []
        self.roiFilter.addItems(roiNames)
        self.roiFilter.currentIndexChanged.connect(self.showRois)
        for i in range(self.roiFilter.count()):
            if currentSel == self.roiFilter.itemText(i):
                self.roiFilter.setCurrentIndex(i)
                break

    def _setRoiSelected(self, roiParam: RoiParams):
        roiParam.selected = True
        roiParam.polygon.set_edgecolor((0, 1, 1, 0.9))  # Highlight selected rois.
        roiParam.polygon.set_linewidth(2)

    def _setAllRoisSelected(self, selected: bool):
        for param in self.rois:
            param.selected = selected
            if selected:
                param.polygon.set_edgecolor((0, 1, 1, 0.9))  # Highlight selected rois.
                param.polygon.set_linewidth(2)
            else:
                param.polygon.set_edgecolor((0, 1, 0, 0.9))
                param.polygon.set_linewidth(1)

    def enableHoverAnnotation(self, enable: bool):
        if enable:
            self._toggleCids = [self._plotWidget.canvas.mpl_connect('motion_notify_event', self._hoverCallback),
                                self._plotWidget.canvas.mpl_connect('button_press_event', self._mouseClickCallback),
                                self._plotWidget.canvas.mpl_connect('key_press_event', self._keyPressCallback),
                                self._plotWidget.canvas.mpl_connect('key_release_event', self._keyReleaseCallback)]
        else:
            if self._toggleCids:
                [self._plotWidget.canvas.mpl_disconnect(cid) for cid in self._toggleCids]

    def showRois(self):
        pattern = self.roiFilter.currentText()
        self._clearRois()
        for name, num, fformat in self.metadata.getRois():
            if re.fullmatch(pattern, name):
                try:
                    self._addPolygonForRoi(self._roiManager.getROI(self.metadata, name, num))
                except Exception as e:
                    logger = logging.getLogger(__name__)
                    logger.warning(f"Failed to load Roi with name: {name}, number: {num}, format: {fformat.name}")
                    logger.exception(e)
        self._plotWidget.canvas.draw_idle()

    # Signal handlers for RoiManager
    def _onRoiRemoved(self, roiFile: pwsdt.RoiFile):  # This is most likely triggered by this widget's own actions, but it could also be external modification of the roiManager
        if self.metadata == roiFile.acquisition:  # ROI belongs to the currently displayed ROI
            self._removePolygonForRoi(roiFile)
            self._plotWidget.canvas.draw_idle()

    def _onRoiUpdated(self, roiFile: pwsdt.RoiFile):
        if self.metadata == roiFile.acquisition:  # ROI belongs to the currently displayed ROI
            self._removePolygonForRoi(roiFile)
            self._addPolygonForRoi(roiFile)
            self._plotWidget.canvas.draw_idle()

    def _onRoiCreated(self, roiFile: pwsdt.RoiFile, mayHaveBeenOverwrite: bool):
        if self.metadata == roiFile.acquisition:  # ROI belongs to the currently displayed ROI
            if mayHaveBeenOverwrite:
                for param in self.rois:
                    if roiFile is param.roiFile:
                        self._removePolygonForRoi(roiFile)
                        break
            self._addPolygonForRoi(roiFile)
            self._plotWidget.canvas.draw_idle()

    def _hoverCallback(self, event):  # Show an annotation about the ROI when the mouse hovers over it.
        def update_annot(roiFile: pwsdt.RoiFile, poly: PathPatch):
            self.annot.xy = poly.get_path().vertices.mean(axis=0)  # Set the location to the center of the polygon.
            text = f"{roiFile.name}, {roiFile.number}"
            if self.metadata.pws:  # A day may come where fluorescence is not taken on the same camera as pws, in this case we will have multiple pixel sizes and ROI handling will need an update. for now just assume we'll use PWS pixel size
                if self.metadata.pws.pixelSizeUm:  # For some systems (nanocytomics) this is None
                    text += f"\n{self.metadata.pws.pixelSizeUm ** 2 * np.sum(roiFile.getRoi().mask):.2f} $μm^2$"
            self.annot.set_text(text)
            self.annot.get_bbox_patch().set_alpha(0.4)

        vis = self.annot.get_visible()  # Is the annotation already being shown?
        if event.inaxes == self._plotWidget.ax:  # The event takes place in the axes of this widget.
            for params in self.rois:
                contained, _ = params.polygon.contains(event)
                if contained:
                    if not vis:  # If we aren't already showing the annotation then show it for the currently hovered ROI.
                        update_annot(params.roiFile, params.polygon)
                        self.annot.set_visible(True)
                        self._plotWidget.canvas.draw_idle()
                    return
            if vis:  # If we got here then no hover actions were found. If an annotation is currently being shown turn off the annotation.
                self.annot.set_visible(False)
                self._plotWidget.canvas.draw_idle()

    def _keyPressCallback(self, event: KeyEvent):
        key = event.key.lower()
        if key == 'a':  # Select/Deselect All
            self._selectAllFunc()
        elif key == 'm':  # Modify ROI
            selected = [param for param in self.rois if param.selected]
            if len(selected) != 1:
                return  # Only can be done for a single selection.
            self._editFunc(selected[0])
        elif key == 's':  # Shift/rotate
            self._moveRoisFunc()

    def _keyReleaseCallback(self, event: KeyEvent):
        pass

    def _mouseClickCallback(self, event: MouseEvent):
        # Determine if a ROI was clicked on
        _ = [param for param in self.rois if param.polygon.contains(event)[0]]
        if len(_) > 0:
            selectedROIParam = _[0]  # There should have only been one roiFile clicked on. select the first one from the list (hopefully only one there anyway)
        else:
            selectedROIParam = None  # No Roi was clicked

        if event.button == 1:  # Left click
            exclusive = not (Qt.ControlModifier & QApplication.keyboardModifiers())  # If control isn't being pressed then change to only select this one ROI. the QApplication method returns a bitmask of Qt.KeyboardModifiers
            if exclusive:
                self._setAllRoisSelected(False)
            if selectedROIParam is not None:  # Didn't click on an ROI
                self._setRoiSelected(selectedROIParam)
            self._plotWidget.canvas.draw_idle()
        if event.button == 3:  # "3" is the right button
            self._showRightClickMenu(selectedROIParam)

    def _addPolygonForRoi(self, roiFile: pwsdt.RoiFile):
        roi = roiFile.getRoi()
        if roi.verts is not None:
            poly: PathPatch = descartes.PolygonPatch(roi.polygon, facecolor=(1, 0, 0, 0.5), linewidth=1, edgecolor=(0, 1, 0, 0.9))
            poly.set_picker(0)  # allow the polygon to trigger a pickevent
            self._plotWidget.ax.add_patch(poly)
            self.rois.append(RoiParams(roiFile, poly, False))

    def _removePolygonForRoi(self, roiFile: pwsdt.RoiFile):
        parm = None
        for param in self.rois:
            if param.roiFile is roiFile:
                param.polygon.remove()
                parm = param
        if parm is None:  # No matching roiParam was found, that ain't right.
            raise ValueError(f"RoiPlot did not find a RoiParam matching RoiFile: {roiFile}")
        else:
            self.rois.remove(parm)

    def _clearRois(self):
        for param in self.rois:
            param.polygon.remove()
        self.rois = []

    def _exportAction(self):
        def showSinCityDlg():
            dlg = SinCityDlg(self, self)
            dlg.show()
        menu = QMenu("Export Menu", parent=self)
        act = QAction("Colored Nuclei")
        act.triggered.connect(showSinCityDlg)
        menu.addAction(act)
        menu.exec(self.mapToGlobal(self.exportButton.pos()))

    def _showRightClickMenu(self, selectedROIParam: RoiParams):
        """

        Args:
         selectedROIParam: The ROI params associated with the ROI that was clicked on.
        """
        # Actions that can happen even if no ROI was clicked on.
        def deleteFunc():
            for param in tuple(self.rois): # Create a tuple copy of the list since self.rois will be modified as we loop which causes weird behavior.
                if param.selected:
                    self._roiManager.removeRoi(param.roiFile)  # Signals emitted here will causes the necesarry UI updates.
                    self.roiDeleted.emit(self.metadata, param.roiFile)

        def copyFunc():
            d = {}
            for param in self.rois:
                if param.selected:
                    roiFile = param.roiFile
                    d[(roiFile.name, roiFile.number)] = roiFile.getRoi()
            mimeData = QMimeData()
            mimeData.setData('pickleRoi', pickle.dumps(d))
            QApplication.clipboard().setMimeData(mimeData)

        def pasteFunc():
            try:
                b = QApplication.clipboard().mimeData().data('pickleRoi')
                d: dict = pickle.loads(b)
                for k, v in d.items():
                    try:
                        roiFile = self._roiManager.createRoi(self.metadata, v, k[0], k[1], overwrite=False)
                        self.roiCreated.emit(self.metadata, roiFile)
                    except OSError:
                        logging.getLogger(__name__).info(f"Attempting to paste and ROI that already exists. Cannot Overwrite. {v.name}, {v.number}")
            except Exception as e:
                QMessageBox.information(self, "Nope", 'Pasting Failed. See the log.')
                logging.getLogger(__name__).exception(e)

        popMenu = QMenu(self)
        deleteAction = popMenu.addAction("Delete Selected ROIs", deleteFunc)
        moveAction = popMenu.addAction("(S)hift/Rotate Selected ROIs", self._moveRoisFunc)
        selectAllAction = popMenu.addAction("De/Select (A)ll", self._selectAllFunc)
        copyAction = popMenu.addAction("Copy ROIs", copyFunc)
        pasteAction = popMenu.addAction("Paste ROIs", pasteFunc)

        if not any([roiParam.selected for roiParam in self.rois]):  # If no rois are selected then some actions can't be performed
            deleteAction.setEnabled(False)
            moveAction.setEnabled(False)
            copyAction.setEnabled(False)

        moveAction.setToolTip(MovingModifier.getHelpText())
        popMenu.setToolTipsVisible(True)

        if selectedROIParam is not None:
            # Actions that require that a ROI was clicked on.
            popMenu.addSeparator()
            popMenu.addAction("(M)odify", lambda sel=selectedROIParam: self._editFunc(sel))

        cursor = QCursor()
        popMenu.popup(cursor.pos())

    def _moveRoisFunc(self):
        """Callback triggered from right click or key press"""
        coordSet = []
        selectedROIParams = []
        for param in self.rois:
            if param.selected:
                selectedROIParams.append(param)
                coordSet.append(param.roiFile.getRoi().verts)
        if len(selectedROIParams) == 0:  # Nothing to do
            return

        def done(vertsSet, handles):
            self._polyWidg.set_active(False)
            self._polyWidg.set_visible(False)
            for param, verts in zip(selectedROIParams, vertsSet):
                newRoi = pwsdt.Roi.fromVerts(np.array(verts),
                                             param.roiFile.getRoi().mask.shape)
                self._roiManager.updateRoi(param.roiFile, newRoi)
                self.roiModified.emit(self.metadata, param.roiFile)

            self.enableHoverAnnotation(True)

        def cancelled():
            self.enableHoverAnnotation(True)

        self.enableHoverAnnotation(False)  # This should be reenabled when the widget is finished or cancelled.
        self._polyWidg = MovingModifier(self.ax, onselect=done, onCancelled=cancelled)
        self._polyWidg.set_active(True)
        self._polyWidg.initialize(coordSet)

    def _selectAllFunc(self):
        """Callback triggered by right click or key press"""
        sel = not any([param.selected for param in self.rois])  # Determine whether to select or deselect all
        self._setAllRoisSelected(sel)
        self._plotWidget.canvas.draw_idle()

    def _editFunc(self, selectedROIParam: RoiParams):
        """Callback triggered by right click or key press, required a selected Roi"""
        # extract handle points from the polygon
        poly = shapelyPolygon(selectedROIParam.roiFile.getRoi().verts)
        poly = poly.buffer(0)
        poly = poly.simplify(poly.length ** .5 / 5, preserve_topology=False)
        handles = poly.exterior.coords

        def done(verts, handles):
            verts = verts[0]
            newRoi = pwsdt.Roi.fromVerts(np.array(verts), selectedROIParam.roiFile.getRoi().mask.shape)
            self._polyWidg.set_active(False)
            self._polyWidg.set_visible(False)
            self._roiManager.updateRoi(selectedROIParam.roiFile, newRoi)
            self.roiModified.emit(self.metadata, selectedROIParam.roiFile)

            self.enableHoverAnnotation(True)

        def cancelled():
            self.enableHoverAnnotation(True)

        self._polyWidg = PolygonModifier(self.ax, onselect=done, onCancelled=cancelled)
        self._polyWidg.set_active(True)
        self.enableHoverAnnotation(False)
        self._polyWidg.initialize([handles])


class WhiteSpaceValidator(QValidator):
    stateChanged = QtCore.pyqtSignal(QValidator.State)

    def __init__(self):
        super().__init__()
        self.state = QValidator.Acceptable

    def validate(self, inp: str, pos: int):
        oldState = self.state
        inp = self.fixup(inp)
        self.state = QValidator.Acceptable
        if self.state != oldState: self.stateChanged.emit(self.state)
        return self.state, inp, pos

    def fixup(self, a0: str) -> str:
        return a0.strip()




import typing

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, QItemSelection, QModelIndex, QItemSelectionModel
from PyQt5.QtWidgets import QTreeView, QWidget, QTreeWidget, QTreeWidgetItem, QAbstractItemView, QAbstractItemDelegate

from pwspy_gui.PWSAnalysisApp.plugins.acquisitionSequencer._treeModel import TreeModel
from pwspy.utility.acquisition import SequencerStep
from pwspy_gui.PWSAnalysisApp.plugins.acquisitionSequencer._ui.Delegate import IterationRangeDelegate
from pwspy.utility.acquisition.sequencerCoordinate import SequencerCoordinateRange


class MyTreeView(QTreeView):
    newCoordSelected = pyqtSignal(SequencerCoordinateRange)
    currentItemChanged = pyqtSignal(SequencerStep)

    def __init__(self, parent: QWidget = None):
        super().__init__(parent=parent)
        delegate = IterationRangeDelegate(self)
        self.setItemDelegate(delegate)
        delegate.editingFinished.connect(lambda: self._selectionChanged(self.selectionModel().selection())) # When we edit an item we still want to process it as a change even though the selection hasn't changed.
        self.setEditTriggers(QAbstractItemView.AllEditTriggers)  # Make editing start on a single click.
        self.setIndentation(10)  # Reduce the default indentation
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)  # Smooth scrolling
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self._currentCoordRange = None

    def commitAndClose(self):
        """Quickly deselect and reselect. this forces any open editor to commit its changes."""
        idx = self.currentIndex()
        self.setCurrentIndex(QModelIndex())
        self.setCurrentIndex(idx)

        delegate: IterationRangeDelegate = self.itemDelegate()
        try:
            self.closeEditor(delegate.editor, QAbstractItemDelegate.NoHint)  # Close the editor to indicate to the user that the change has been accepted.
        except RuntimeError:  # Sometimes the editor has been deleted in the c++ layer causing an issue
            pass

    def setRoot(self, root: SequencerStep) -> None:
        """
        Populate the widget with a sequence of acquisition steps

        Args:
            root: The Root step of the acquisition sequence. All other steps are children of this step.
        """
        self.setModel(TreeModel(root))
        self.setSelectionModel(QItemSelectionModel(self.model(), self))
        self.selectionModel().selectionChanged.connect(self._selectionChanged)
        self.selectionModel().currentChanged.connect(self._currentChanged)

    def _selectionChanged(self, selected: QItemSelection, deselected: QItemSelection = None):
        try:
            idx = selected.indexes()[0]  # We only support a single selection anyways.
        except IndexError:
            return  # Sometime this can get fired with no selected indexes.
        step: SequencerStep = idx.internalPointer()
        coordSteps = []
        while step is not self.model().invisibleRootItem(): # This will break out once we reach the root item.
            coordStep = step.data(QtCore.Qt.EditRole)  # The item delegate saves an tuple indicated the range of iterations selected in the `editRole` data slot of steps.
            if coordStep is None:
                coordSteps.append((step.id, None))
            else:
                coordSteps.append(coordStep)
            step = step.parent()  # On the next iteration look at the parent of the selected step.
        self._currentCoordRange = SequencerCoordinateRange(list(reversed(coordSteps)))
        self.newCoordSelected.emit(self._currentCoordRange)

    def getCurrentSelectedCoordinateRange(self):
        return self._currentCoordRange

    def _currentChanged(self, current: QModelIndex, previous: QModelIndex):
        if current.internalPointer() is not None:  # In some cases we may change to index to something blank that has not pointer
            self.currentItemChanged.emit(current.internalPointer())


class DictTreeView(QTreeWidget):
    """
    A QTreeWidget that displays that contents of a Python `dict`
    Read-Only.
    """

    def setDict(self, d: dict) -> None:
        """
        Set the `dict` that this widget displays.

        Args:
            d: The dictionary to display the contents of.
        """
        self.clear()
        self._fillItem(self.invisibleRootItem(), d)

    @staticmethod
    def _fillItem(item: QTreeWidgetItem, value: typing.Union[dict, list]):
        """Recursively populate a tree item with children to match the contents of a `dict`"""
        item.setExpanded(True)
        if isinstance(value, dict):
            for key, val in value.items():
                child = QTreeWidgetItem()
                child.setText(0, f"{key}")
                item.addChild(child)
                if isinstance(val, (list, dict)):
                    DictTreeView._fillItem(child, val)
                else:
                    child.setText(1, f"{val}")
        elif isinstance(value, list):
            for val in value:
                child = QTreeWidgetItem()
                item.addChild(child)
                if type(val) is dict:
                    child.setText(0, '[dict]')
                    DictTreeView._fillItem(child, val)
                elif type(val) is list:
                    child.setText(0, '[list]')
                    DictTreeView._fillItem(child, val)
                else:
                    child.setText(0, val)
                    child.setExpanded(True)





# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import os

import numpy as np
from PyQt5 import QtCore
from PyQt5.QtCore import QPoint
from PyQt5.QtGui import QPixmap, QImage
from PyQt5.QtWidgets import QMenu, QAction, QWidget, QLabel, QVBoxLayout, QApplication
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateAnalysisResults
from pwspy.dataTypes import Acquisition, PwsCube
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._widgets import AnalysisPlotter
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._analysisViewer import AnalysisViewer
from mpl_qt_viz.visualizers import PlotNd


class LittlePlot(AnalysisPlotter, QWidget):
    def __init__(self, acquisition: Acquisition, analysis: ConglomerateAnalysisResults, title: str, text: str = None,
                 initialField=AnalysisPlotter.PlotFields.Thumbnail):
        assert analysis is not None #The member of the conglomerateAnalysisResults can be None but the way this class is written requires that the object itself exists.
        AnalysisPlotter.__init__(self, acquisition, analysis)
        QWidget.__init__(self)
        self.setLayout(QVBoxLayout())
        self.titleLabel = QLabel(title, self)
        self.titleLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.imLabel = QLabel(self)
        self.imLabel.setScaledContents(True)
        self.layout().addWidget(self.titleLabel)
        if text is not None:
            self.textLabel = QLabel(self)
            self.textLabel.setStyleSheet("QLabel {color: #b40000}") #This isn't working for some reason
            self.textLabel.setText(text)
            self.textLabel.setAlignment(QtCore.Qt.AlignCenter)
            self.layout().addWidget(self.textLabel)
        self.layout().addWidget(self.imLabel)
        self.title = title
        self.setMinimumWidth(20)
        self.changeData(initialField)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)
        self.plotnd = None #Just a reference to a plotND class instance so it isn't deleted.

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            mainWindow = QApplication.instance().window
            viewer = AnalysisViewer(metadata=self.acq, analysisLoader=self.analysis, title=self.title, roiManager=QApplication.instance().roiManager, parent=mainWindow, initialField=self.analysisField, flags=QtCore.Qt.Window)
            viewer.show()

    def changeData(self, field: AnalysisPlotter.PlotFields):
        AnalysisPlotter.changeData(self, field)
        data = self.data
        data = data - np.percentile(data, 0.1)
        data = (data / np.percentile(data, 99.9) * 255)
        data[data < 0] = 0
        data[data > 255] = 255
        data = data.astype(np.uint8)
        p = QPixmap.fromImage(QImage(data.data, data.shape[1], data.shape[0], data.strides[0], QImage.Format_Grayscale8))
        self.imLabel.setPixmap(p)

    def showContextMenu(self, point: QPoint):
        menu = QMenu("ContextMenu", self)
        menu.setToolTipsVisible(True)
        if self.analysis.pws is not None:
            anPlotAction = QAction("Plot PWS Analyzed Reflectance", self)
            anPlotAction.triggered.connect(self.plotAn3d)
            menu.addAction(anPlotAction)
            if 'reflectance' in self.analysis.pws.file.keys():
                opdAction = QAction("Plot OPD", self)
                opdAction.triggered.connect(self.plotOpd3d)
                menu.addAction(opdAction)
                rmsDepthAction = QAction("Plot RMS-by-OPD", self)
                rmsDepthAction.triggered.connect(self.plotRMSOpd3d)
                rmsDepthAction.setToolTip(self.plotRMSOpd3d.__doc__)
                menu.addAction(rmsDepthAction)
        if self.acq.pws is not None:
            rawPlotAction = QAction("Plot PWS Raw Data", self)
            rawPlotAction.triggered.connect(self.plotRaw3d)
            menu.addAction(rawPlotAction)
        if self.analysis.dyn is not None:
            dynAnPlotAction = QAction("Plot DYN Analyzed Reflectance", self)
            dynAnPlotAction.triggered.connect(self.plotDynAn3d)
            menu.addAction(dynAnPlotAction)
        if self.acq.dynamics is not None:
            dynRawPlotAction = QAction("Plot DYN Raw Data", self)
            dynRawPlotAction.triggered.connect(self.plotDynRaw3d)
            menu.addAction(dynRawPlotAction)
        menu.exec(self.mapToGlobal(point))

    def plotAn3d(self):
        refl = self.analysis.pws.reflectance
        refl += self.analysis.pws.meanReflectance[:, :, None]  # The `reflectance has had the mean subtracted. add it back in.
        self.plotnd = PlotNd(refl.data, title=os.path.split(self.acq.filePath)[-1], names=('y', 'x', 'k (rad/um)'),
                             indices=[range(refl.data.shape[0]), range(refl.data.shape[1]), refl.wavenumbers], parent=self)

    def plotRaw3d(self):
        im = PwsCube.fromMetadata(self.acq.pws)
        self.plotnd = PlotNd(im.data, title=os.path.split(self.acq.filePath)[-1], names=('y', 'x', 'lambda'),
                             indices=[range(im.data.shape[0]), range(im.data.shape[1]), im.wavelengths], parent=self)

    def plotOpd3d(self):
        opd, opdIndex = self.analysis.pws.opd
        self.plotnd = PlotNd(opd, names=('y', 'x', 'um'), title=os.path.split(self.acq.filePath)[-1],
                             indices=[range(opd.shape[0]), range(opd.shape[1]), opdIndex], parent=self)

    def plotRMSOpd3d(self):
        """
        Square root of cumulative sum of OPD^2. Important note: The units are in terms of OPD not depth. Since the light makes a round trip -> depth = OPD / (2 * RI of cell)
        """
        opd: np.ndarray
        opd, opdIndx = self.analysis.pws.opd

        opdSquaredSum = np.cumsum(opd**2, axis=2)  # Parseval's theorem tells us that this is equivalent to the sum of the squares of our original signal. Cumulative sum from 0 up to a given OPD
        opdSquaredSum *= len(self.analysis.pws.reflectance.wavenumbers) / opd.shape[2]  # If the original data and opd were of the same length then the above line would be correct. Since the fft may have been upsampled. we need to normalize.
        rmsByOPD = np.sqrt(opdSquaredSum)

        self.plotnd = PlotNd(rmsByOPD, names=('y', 'x', 'um'), title=os.path.split(self.acq.filePath)[-1],
                             indices=[range(opd.shape[0]), range(opd.shape[1]), opdIndx], parent=self)

    def plotDynAn3d(self):
        refl = self.analysis.dyn.reflectance
        self.plotnd = PlotNd(refl.data, title=os.path.split(self.acq.filePath)[-1],
                             names=('y', 'x', 't'),
                             indices=[range(refl.data.shape[0]), range(refl.data.shape[1]), refl.times], parent=self)

    def plotDynRaw3d(self):
        im = self.acq.dynamics.toDataClass()
        self.plotnd = PlotNd(im.data, title=os.path.split(self.acq.filePath)[-1], names=('y', 'x', 't'),
                             indices=[range(im.data.shape[0]), range(im.data.shape[1]), im.times], parent=self)


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
GUI applications used for analysis and calibration of PWS (Partial Wave Spectroscopic) microscopy data.

PWSAnalysisApp
----------------

The main application used for the analysis of PWS and related data.

ExtraReflectanceCreator
-------------------------

This application is used to generate `ExtraReflectanceCube` calibration files and upload them to google drive.

"""
import logging
try:
    from . import version
    _versionStr = version.version
except ImportError:  # When running from source the version.py file won't have been created by setuptools_scm
    logging.getLogger(__name__).info("Failed to import `version.py`. Trying to import setuptools_scm")
    try:
        import setuptools_scm
        _versionStr = setuptools_scm.get_version(root='../..', relative_to=__file__)
    except ImportError:  # setuptools_scm
        logging.getLogger(__name__).info("Failed to import setuptools_scm. Using fallback version string.")
        _versionStr = "x.x.x-dev"

__author__ = "Nick Anthony"
__version__ = _versionStr

__all__ = ['resources', 'appPath']
import os

resources = os.path.join(os.path.split(__file__)[0], '_resources')

appPath = os.path.expanduser('~/PwspyApps') # Create a directory to store all application data
if not os.path.exists(appPath):
    os.mkdir(appPath)



import dataclasses
import pwspy.dataTypes as pwsdt
from abc import ABC, abstractmethod
import typing
from pwspy.analysis import AbstractAnalysisSettings
from pwspy.analysis.dynamics import DynamicsAnalysisSettings
from pwspy.analysis.pws import PWSAnalysisSettings
from pwspy.dataTypes import MetaDataBase


class AbstractRuntimeAnalysisSettings(ABC):
    """This represents all the information that gets passed to an analysis.
    Unlike AnalysisSettings they can contain objects which are not meant to be saved/loaded.
    This includes the references to the acquisitions to be analyzed as well as the files used
    for calibration and normalization"""

    @abstractmethod
    def getSaveableSettings(self) -> AbstractAnalysisSettings:
        """

        Returns:
            Only the settings which can be saved.
        """
        pass

    @abstractmethod
    def getAnalysisName(self) -> str:
        """

        Returns:
            The name the analysis is referred to by
        """
        pass

    @abstractmethod
    def getReferenceMetadata(self) -> MetaDataBase:
        """

        Returns:
            The metadata object referring to the reference image used for normalization.
        """
        pass

    @abstractmethod
    def getCellMetadatas(self) -> typing.Sequence[MetaDataBase]:
        """

        Returns:
            A sequence of metadata objects for the acquisitions to be analyzed.
        """
        pass

    @abstractmethod
    def getExtraReflectanceMetadata(self) -> typing.Optional[pwsdt.ERMetaData]:
        """

        Returns:
            The metadata object for the extra reflection correction. This can be none if you want to skip this correction.
        """
        pass


@dataclasses.dataclass
class DynamicsRuntimeAnalysisSettings(AbstractRuntimeAnalysisSettings):  # Inherit docstring
    settings: DynamicsAnalysisSettings
    extraReflectanceMetadata: typing.Optional[pwsdt.ERMetaData]
    referenceMetadata: pwsdt.DynMetaData
    cellMetadata: typing.List[pwsdt.DynMetaData]
    analysisName: str

    def getSaveableSettings(self) -> DynamicsAnalysisSettings:  # Inherit docstring
        return self.settings

    def getAnalysisName(self) -> str:
        return self.analysisName

    def getReferenceMetadata(self) -> pwsdt.DynMetaData:
        return self.referenceMetadata

    def getCellMetadatas(self) -> typing.Sequence[pwsdt.DynMetaData]:
        return self.cellMetadata

    def getExtraReflectanceMetadata(self) -> pwsdt.ERMetaData:
        return self.extraReflectanceMetadata


@dataclasses.dataclass
class PWSRuntimeAnalysisSettings(AbstractRuntimeAnalysisSettings):  # Inherit docstring
    settings: PWSAnalysisSettings
    extraReflectanceMetadata: typing.Optional[pwsdt.ERMetaData]
    referenceMetadata: pwsdt.ICMetaData
    cellMetadata: typing.List[pwsdt.ICMetaData]
    analysisName: str

    def getSaveableSettings(self) -> PWSAnalysisSettings:
        return self.settings

    def getAnalysisName(self) -> str:
        return self.analysisName

    def getReferenceMetadata(self) -> pwsdt.ICMetaData:
        return self.referenceMetadata

    def getCellMetadatas(self) -> typing.Sequence[pwsdt.ICMetaData]:
        return self.cellMetadata

    def getExtraReflectanceMetadata(self) -> pwsdt.ERMetaData:
        return self.extraReflectanceMetadata
